# pylint: disable=line-too-long

""" StructureToPython class for converting JSON Structure schema to Python classes """

import json
import os
import re
import random
from typing import Any, Dict, List, Set, Tuple, Union, Optional

from avrotize.common import pascal, process_template
from avrotize.jstructtoavro import JsonStructureToAvro

JsonNode = Dict[str, 'JsonNode'] | List['JsonNode'] | str | None

INDENT = '    '


def is_python_reserved_word(word: str) -> bool:
    """Checks if a word is a Python reserved word"""
    reserved_words = [
        'False', 'None', 'True', 'and', 'as', 'assert', 'async', 'await',
        'break', 'class', 'continue', 'def', 'del', 'elif', 'else', 'except',
        'finally', 'for', 'from', 'global', 'if', 'import', 'in', 'is',
        'lambda', 'nonlocal', 'not', 'or', 'pass', 'raise', 'return',
        'try', 'while', 'with', 'yield', 'record', 'self', 'cls'
    ]
    return word in reserved_words


class StructureToPython:
    """ Converts JSON Structure schema to Python classes """

    def __init__(self, base_package: str = '', dataclasses_json_annotation=False, avro_annotation=False) -> None:
        self.base_package = base_package
        self.dataclasses_json_annotation = dataclasses_json_annotation
        self.avro_annotation = avro_annotation
        self.output_dir = os.getcwd()
        self.schema_doc: JsonNode = None
        self.generated_types: Dict[str, str] = {}
        self.generated_structure_types: Dict[str, Dict[str, Union[str, Dict, List]]] = {}
        self.type_dict: Dict[str, Dict] = {}
        self.definitions: Dict[str, Any] = {}
        self.schema_registry: Dict[str, Dict] = {}

    def get_qualified_name(self, namespace: str, name: str) -> str:
        """ Concatenates namespace and name with a dot separator """
        return f"{namespace}.{name}" if namespace != '' else name

    def concat_namespace(self, namespace: str, name: str) -> str:
        """ Concatenates namespace and name with a dot separator """
        if namespace and name:
            return f"{namespace}.{name}"
        elif namespace:
            return namespace
        else:
            return name

    def map_primitive_to_python(self, structure_type: str) -> str:
        """ Maps JSON Structure primitive types to Python types """
        mapping = {
            'null': 'None',
            'boolean': 'bool',
            'string': 'str',
            'integer': 'int',
            'number': 'float',
            'int8': 'int',
            'uint8': 'int',
            'int16': 'int',
            'uint16': 'int',
            'int32': 'int',
            'uint32': 'int',
            'int64': 'int',
            'uint64': 'int',
            'int128': 'int',
            'uint128': 'int',
            'float8': 'float',
            'float': 'float',
            'double': 'float',
            'binary32': 'float',
            'binary64': 'float',
            'decimal': 'decimal.Decimal',
            'binary': 'bytes',
            'date': 'datetime.date',
            'time': 'datetime.time',
            'datetime': 'datetime.datetime',
            'timestamp': 'datetime.datetime',
            'duration': 'datetime.timedelta',
            'uuid': 'uuid.UUID',
            'uri': 'str',
            'jsonpointer': 'str',
            'any': 'typing.Any'
        }
        qualified_class_name = self.get_qualified_name(
            self.base_package.lower(), structure_type.lower())
        if qualified_class_name in self.generated_types:
            result = qualified_class_name
        else:
            result = mapping.get(structure_type, 'typing.Any')
        return result

    def is_python_primitive(self, type_name: str) -> bool:
        """ Checks if a type is a Python primitive type """
        return type_name in ['None', 'bool', 'int', 'float', 'str', 'bytes']

    def is_python_typing_struct(self, type_name: str) -> bool:
        """ Checks if a type is a Python typing type """
        return type_name.startswith('typing.Dict[') or type_name.startswith('typing.List[') or \
               type_name.startswith('typing.Optional[') or type_name.startswith('typing.Union[') or \
               type_name == 'typing.Any'

    def safe_name(self, name: str) -> str:
        """Converts a name to a safe Python name"""
        if is_python_reserved_word(name):
            return name + "_"
        return name

    def pascal_type_name(self, ref: str) -> str:
        """Converts a reference to a type name"""
        return '_'.join([pascal(part) for part in ref.split('.')[-1].split('_')])

    def python_package_from_structure_type(self, namespace: str, type_name: str) -> str:
        """Gets the Python package from a type name"""
        type_name_package = '.'.join([part.lower() for part in type_name.split('.')]) if '.' in type_name else type_name.lower()
        if '.' in type_name:
            package = type_name_package
        else:
            namespace_package = '.'.join([part.lower() for part in namespace.split('.')]) if namespace else ''
            package = namespace_package + ('.' if namespace_package and type_name_package else '') + type_name_package
        if self.base_package:
            package = self.base_package + '.' + package
        return package

    def python_type_from_structure_type(self, type_name: str) -> str:
        """Gets the Python class from a type name"""
        return self.pascal_type_name(type_name)

    def python_fully_qualified_name_from_structure_type(self, namespace: str, type_name: str) -> str:
        """Gets the fully qualified Python class name from a Structure type."""
        package = self.python_package_from_structure_type(namespace, type_name)
        return package + ('.' if package else '') + self.python_type_from_structure_type(type_name)

    def strip_package_from_fully_qualified_name(self, fully_qualified_name: str) -> str:
        """Strips the package from a fully qualified name"""
        return fully_qualified_name.split('.')[-1]

    def resolve_ref(self, ref: str, context_schema: Optional[Dict] = None) -> Optional[Dict]:
        """ Resolves a $ref to the actual schema definition """
        if not ref.startswith('#/'):
            if ref in self.schema_registry:
                return self.schema_registry[ref]
            return None

        path = ref[2:].split('/')
        schema = context_schema if context_schema else self.schema_doc
        for part in path:
            if not isinstance(schema, dict) or part not in schema:
                return None
            schema = schema[part]
        return schema

    def register_schema_ids(self, schema: Dict, base_uri: str = '') -> None:
        """ Recursively registers schemas with $id keywords """
        if not isinstance(schema, dict):
            return

        if '$id' in schema:
            schema_id = schema['$id']
            if base_uri and not schema_id.startswith(('http://', 'https://', 'urn:')):
                from urllib.parse import urljoin
                schema_id = urljoin(base_uri, schema_id)
            self.schema_registry[schema_id] = schema
            base_uri = schema_id

        if 'definitions' in schema:
            for def_name, def_schema in schema['definitions'].items():
                if isinstance(def_schema, dict):
                    self.register_schema_ids(def_schema, base_uri)

        if 'properties' in schema:
            for prop_name, prop_schema in schema['properties'].items():
                if isinstance(prop_schema, dict):
                    self.register_schema_ids(prop_schema, base_uri)

        for key in ['items', 'values', 'additionalProperties']:
            if key in schema and isinstance(schema[key], dict):
                self.register_schema_ids(schema[key], base_uri)

    def convert_structure_type_to_python(self, class_name: str, field_name: str, 
                                        structure_type: JsonNode, parent_namespace: str, 
                                        import_types: Set[str]) -> str:
        """ Converts JSON Structure type to Python type """
        if isinstance(structure_type, str):
            python_type = self.map_primitive_to_python(structure_type)
            if python_type.startswith('datetime.') or python_type == 'decimal.Decimal' or python_type == 'uuid.UUID':
                import_types.add(python_type)
            return python_type
        elif isinstance(structure_type, list):
            # Handle type unions
            non_null_types = [t for t in structure_type if t != 'null']
            if len(non_null_types) == 1:
                inner_type = self.convert_structure_type_to_python(
                    class_name, field_name, non_null_types[0], parent_namespace, import_types)
                if 'null' in structure_type:
                    return f'typing.Optional[{inner_type}]'
                return inner_type
            else:
                union_types = [self.convert_structure_type_to_python(
                    class_name, field_name, t, parent_namespace, import_types) for t in non_null_types]
                return f"typing.Union[{', '.join(union_types)}]"
        elif isinstance(structure_type, dict):
            # Handle $ref
            if '$ref' in structure_type:
                ref_schema = self.resolve_ref(structure_type['$ref'], self.schema_doc)
                if ref_schema:
                    ref_path = structure_type['$ref'].split('/')
                    type_name = ref_path[-1]
                    ref_namespace = '.'.join(ref_path[2:-1]) if len(ref_path) > 3 else parent_namespace
                    ref = self.generate_class_or_choice(ref_schema, ref_namespace, write_file=True, explicit_name=type_name)
                    import_types.add(ref)
                    return self.strip_package_from_fully_qualified_name(ref)
                return 'typing.Any'

            # Handle enum keyword
            if 'enum' in structure_type:
                enum_ref = self.generate_enum(structure_type, field_name, parent_namespace, write_file=True)
                import_types.add(enum_ref)
                return self.strip_package_from_fully_qualified_name(enum_ref)

            # Handle type keyword
            if 'type' not in structure_type:
                return 'typing.Any'

            struct_type = structure_type['type']

            # Handle complex types
            if struct_type == 'object':
                class_ref = self.generate_class(structure_type, parent_namespace, write_file=True)
                import_types.add(class_ref)
                return self.strip_package_from_fully_qualified_name(class_ref)
            elif struct_type == 'array':
                items_type = self.convert_structure_type_to_python(
                    class_name, field_name+'List', structure_type.get('items', {'type': 'any'}), 
                    parent_namespace, import_types)
                return f"typing.List[{items_type}]"
            elif struct_type == 'set':
                items_type = self.convert_structure_type_to_python(
                    class_name, field_name+'Set', structure_type.get('items', {'type': 'any'}), 
                    parent_namespace, import_types)
                return f"typing.Set[{items_type}]"
            elif struct_type == 'map':
                values_type = self.convert_structure_type_to_python(
                    class_name, field_name+'Map', structure_type.get('values', {'type': 'any'}), 
                    parent_namespace, import_types)
                return f"typing.Dict[str, {values_type}]"
            elif struct_type == 'choice':
                # Generate choice returns a Union type and populates import_types with the choice types
                return self.generate_choice(structure_type, parent_namespace, write_file=True, import_types=import_types)
            elif struct_type == 'tuple':
                tuple_ref = self.generate_tuple(structure_type, parent_namespace, write_file=True)
                import_types.add(tuple_ref)
                return self.strip_package_from_fully_qualified_name(tuple_ref)
            else:
                return self.convert_structure_type_to_python(class_name, field_name, struct_type, parent_namespace, import_types)
        return 'typing.Any'

    def generate_class_or_choice(self, structure_schema: Dict, parent_namespace: str, 
                                 write_file: bool = True, explicit_name: str = '') -> str:
        """ Generates a Class or Choice """
        struct_type = structure_schema.get('type', 'object')
        if struct_type == 'object':
            return self.generate_class(structure_schema, parent_namespace, write_file, explicit_name=explicit_name)
        elif struct_type == 'choice':
            return self.generate_choice(structure_schema, parent_namespace, write_file, explicit_name=explicit_name)
        elif struct_type == 'tuple':
            return self.generate_tuple(structure_schema, parent_namespace, write_file, explicit_name=explicit_name)
        return 'typing.Any'

    def generate_class(self, structure_schema: Dict, parent_namespace: str, 
                      write_file: bool, explicit_name: str = '') -> str:
        """ Generates a Python dataclass from JSON Structure object type """
        import_types: Set[str] = set()

        # Get name and namespace
        class_name = pascal(explicit_name if explicit_name else structure_schema.get('name', 'UnnamedClass'))
        schema_namespace = structure_schema.get('namespace', parent_namespace)
        namespace = self.concat_namespace(self.base_package, schema_namespace).lower()
        python_qualified_name = self.python_fully_qualified_name_from_structure_type(schema_namespace, class_name)
        
        if python_qualified_name in self.generated_types:
            return python_qualified_name

        # Check if this is an abstract type
        is_abstract = structure_schema.get('abstract', False)

        # Handle inheritance ($extends)
        base_class = None
        if '$extends' in structure_schema:
            base_ref = structure_schema['$extends']
            if isinstance(self.schema_doc, dict):
                base_schema = self.resolve_ref(base_ref, self.schema_doc)
                if base_schema:
                    ref_path = base_ref.split('/')
                    base_name = ref_path[-1]
                    ref_namespace = '.'.join(ref_path[2:-1]) if len(ref_path) > 3 else parent_namespace
                    base_class = self.generate_class(base_schema, ref_namespace, write_file=True, explicit_name=base_name)
                    import_types.add(base_class)

        # Generate properties
        properties = structure_schema.get('properties', {})
        required_props = structure_schema.get('required', [])

        fields = []
        for prop_name, prop_schema in properties.items():
            field_def = self.generate_field(prop_name, prop_schema, class_name, schema_namespace, 
                                           required_props, import_types)
            fields.append(field_def)

        # Get docstring
        doc = structure_schema.get('description', structure_schema.get('doc', class_name))

        # Generate field docstrings
        field_docstrings = [{
            'name': self.safe_name(field['name']),
            'original_name': field['name'],
            'type': field['type'],
            'is_primitive': field['is_primitive'],
            'is_enum': field['is_enum'],
            'docstring': self.generate_field_docstring(field, schema_namespace),
            'test_value': self.generate_test_value(field),
            'source_type': field.get('source_type', 'string'),
        } for field in fields]

        # If avro_annotation is enabled, convert JSON Structure schema to Avro schema
        # This is embedded in the generated class for runtime Avro serialization
        avro_schema_json = ''
        if self.avro_annotation:
            # Use JsonStructureToAvro to convert the schema
            converter = JsonStructureToAvro()
            schema_copy = structure_schema.copy()
            avro_schema = converter.convert(schema_copy)
            avro_schema_json = json.dumps(avro_schema).replace('\\"', '\'').replace('"', '\\"')

        # Process template
        class_definition = process_template(
            "structuretopython/dataclass_core.jinja",
            class_name=class_name,
            docstring=doc,
            fields=field_docstrings,
            import_types=import_types,
            base_package=self.base_package,
            dataclasses_json_annotation=self.dataclasses_json_annotation,
            avro_annotation=self.avro_annotation,
            avro_schema_json=avro_schema_json,
            is_abstract=is_abstract,
            base_class=base_class,
        )

        if write_file:
            self.write_to_file(namespace, class_name, class_definition)
            self.generate_test_class(namespace, class_name, field_docstrings, import_types)

        self.generated_types[python_qualified_name] = 'class'
        self.generated_structure_types[python_qualified_name] = structure_schema
        return python_qualified_name

    def generate_field(self, prop_name: str, prop_schema: Dict, class_name: str, 
                      parent_namespace: str, required_props: List, import_types: Set[str]) -> Dict:
        """ Generates a field for a Python dataclass """
        field_name = prop_name

        # Check if this is a const field
        if 'const' in prop_schema:
            # Const fields are treated as class variables with default values
            prop_type = self.convert_structure_type_to_python(
                class_name, field_name, prop_schema, parent_namespace, import_types)
            return {
                'name': field_name,
                'type': prop_type,
                'is_primitive': self.is_python_primitive(prop_type) or self.is_python_typing_struct(prop_type),
                'is_enum': False,
                'is_const': True,
                'const_value': prop_schema['const'],
                'source_type': prop_schema.get('type', 'string')
            }

        # Determine if required
        is_required = prop_name in required_props if not isinstance(required_props, list) or \
                     len(required_props) == 0 or not isinstance(required_props[0], list) else \
                     any(prop_name in req_set for req_set in required_props)

        # Get property type
        prop_type = self.convert_structure_type_to_python(
            class_name, field_name, prop_schema, parent_namespace, import_types)

        # Add Optional if not required
        if not is_required and not prop_type.startswith('typing.Optional['):
            prop_type = f'typing.Optional[{prop_type}]'

        # Get source type from structure schema
        source_type = prop_schema.get('type', 'string') if isinstance(prop_schema.get('type'), str) else 'object'

        return {
            'name': field_name,
            'type': prop_type,
            'is_primitive': self.is_python_primitive(prop_type) or self.is_python_typing_struct(prop_type),
            'is_enum': prop_type in self.generated_types and self.generated_types[prop_type] == 'enum',
            'is_const': False,
            'source_type': source_type
        }

    def generate_field_docstring(self, field: Dict, parent_namespace: str) -> str:
        """Generates a field docstring for a Python dataclass"""
        field_type = field['type']
        field_name = self.safe_name(field['name'])
        field_docstring = f"{field_name} ({field_type})"
        return field_docstring

    def generate_enum(self, structure_schema: Dict, field_name: str, parent_namespace: str, 
                     write_file: bool) -> str:
        """ Generates a Python enum from JSON Structure enum """
        # Generate enum name from field name if not provided
        class_name = pascal(structure_schema.get('name', field_name + 'Enum'))
        schema_namespace = structure_schema.get('namespace', parent_namespace)
        namespace = self.concat_namespace(self.base_package, schema_namespace).lower()
        python_qualified_name = self.python_fully_qualified_name_from_structure_type(schema_namespace, class_name)
        
        if python_qualified_name in self.generated_types:
            return python_qualified_name

        symbols = [symbol if not is_python_reserved_word(symbol) else symbol + "_" 
                  for symbol in structure_schema.get('enum', [])]

        doc = structure_schema.get('description', structure_schema.get('doc', f'A {class_name} enum.'))

        enum_definition = process_template(
            "structuretopython/enum_core.jinja",
            class_name=class_name,
            docstring=doc,
            symbols=symbols,
        )

        if write_file:
            self.write_to_file(namespace, class_name, enum_definition)
            self.generate_test_enum(namespace, class_name, symbols)

        self.generated_types[python_qualified_name] = 'enum'
        return python_qualified_name

    def generate_choice(self, structure_schema: Dict, parent_namespace: str, 
                       write_file: bool, explicit_name: str = '', import_types: Optional[Set[str]] = None) -> str:
        """ Generates a Python Union type from JSON Structure choice """
        choice_name = explicit_name if explicit_name else structure_schema.get('name', 'UnnamedChoice')
        schema_namespace = structure_schema.get('namespace', parent_namespace)
        if import_types is None:
            import_types = set()
        
        # If the choice extends a base class, generate the base and derived classes first
        if '$extends' in structure_schema:
            base_ref = structure_schema['$extends']
            if isinstance(self.schema_doc, dict):
                base_schema = self.resolve_ref(base_ref, self.schema_doc)
                if base_schema:
                    # Generate the base class
                    ref_path = base_ref.split('/')
                    base_name = ref_path[-1]
                    ref_namespace = '.'.join(ref_path[2:-1]) if len(ref_path) > 3 else parent_namespace
                    self.generate_class(base_schema, ref_namespace, write_file=True, explicit_name=base_name)
        
        # Generate types for each choice
        choice_types = []
        choices = structure_schema.get('choices', {})
        
        for choice_key, choice_schema in choices.items():
            if isinstance(choice_schema, dict):
                if '$ref' in choice_schema:
                    # Resolve reference and generate the type
                    ref_schema = self.resolve_ref(choice_schema['$ref'], self.schema_doc if isinstance(self.schema_doc, dict) else None)
                    if ref_schema:
                        ref_path = choice_schema['$ref'].split('/')
                        ref_name = ref_path[-1]
                        ref_namespace = '.'.join(ref_path[2:-1]) if len(ref_path) > 3 else parent_namespace
                        qualified_name = self.generate_class(ref_schema, ref_namespace, write_file=True, explicit_name=ref_name)
                        import_types.add(qualified_name)
                        choice_types.append(qualified_name.split('.')[-1])
                elif 'type' in choice_schema:
                    # Generate inline type
                    python_type = self.convert_structure_type_to_python(choice_name, choice_key, choice_schema, schema_namespace, import_types)
                    choice_types.append(python_type)
        
        # Return Union type
        if len(choice_types) == 0:
            return 'typing.Any'
        elif len(choice_types) == 1:
            return choice_types[0]
        else:
            return f"typing.Union[{', '.join(choice_types)}]"

    def generate_tuple(self, structure_schema: Dict, parent_namespace: str, 
                      write_file: bool, explicit_name: str = '') -> str:
        """ Generates a Python Tuple type from JSON Structure tuple """
        # For now, return typing.Any as tuples need special handling
        return 'typing.Any'

    def generate_map_alias(self, structure_schema: Dict, parent_namespace: str,
                          write_file: bool) -> str:
        """ Generates a Python TypeAlias for a top-level map type """
        import_types: Set[str] = set()
        
        # Get name and namespace
        class_name = pascal(structure_schema.get('name', 'UnnamedMap'))
        schema_namespace = structure_schema.get('namespace', parent_namespace)
        namespace = self.concat_namespace(self.base_package, schema_namespace).lower()
        python_qualified_name = self.python_fully_qualified_name_from_structure_type(schema_namespace, class_name)
        
        if python_qualified_name in self.generated_types:
            return python_qualified_name
        
        # Get the value type
        values_schema = structure_schema.get('values', {'type': 'any'})
        values_type = self.convert_structure_type_to_python(
            class_name, 'Values', values_schema, schema_namespace, import_types)
        
        # Get docstring
        doc = structure_schema.get('description', structure_schema.get('doc', f'A {class_name} map type.'))
        
        # Generate the type alias module
        map_definition = process_template(
            "structuretopython/map_alias.jinja",
            class_name=class_name,
            docstring=doc,
            values_type=values_type,
            import_types=import_types,
            base_package=self.base_package
        )
        
        if write_file:
            self.write_to_file(namespace, class_name, map_definition)
        
        self.generated_types[python_qualified_name] = 'map'
        return python_qualified_name

    def generate_test_value(self, field: Dict) -> Any:
        """Generates a test value for a given field"""
        field_type = field['type']

        def generate_value(field_type: str):
            test_values = {
                'str': chr(39) + ''.join([chr(random.randint(97, 122)) for _ in range(0, 20)]) + chr(39),
                'bool': str(random.choice([True, False])),
                'int': f'int({random.randint(0, 100)})',
                'float': f'float({random.uniform(0, 100)})',
                'bytes': 'b"test_bytes"',
                'None': 'None',
                'datetime.date': 'datetime.date.today()',
                'datetime.datetime': 'datetime.datetime.now(datetime.timezone.utc)',
                'datetime.time': 'datetime.datetime.now(datetime.timezone.utc).time()',
                'decimal.Decimal': f'decimal.Decimal("{random.randint(0, 100)}.{random.randint(0, 100)}")',
                'datetime.timedelta': 'datetime.timedelta(days=1)',
                'uuid.UUID': 'uuid.uuid4()',
                'typing.Any': '{"test": "test"}'
            }

            def resolve(field_type: str) -> str:
                pattern = re.compile(r'^(?:typing\.)*(Optional|List|Dict|Union|Set)\[(.+)\]$')
                match = pattern.match(field_type)
                if not match:
                    return field_type
                outer_type, inner_type = match.groups()
                if outer_type == 'Optional':
                    return inner_type
                elif outer_type in ['List', 'Set']:
                    return resolve(inner_type)
                elif outer_type == 'Dict':
                    _, value_type = inner_type.split(',', 1)
                    return resolve(value_type.strip())
                elif outer_type == 'Union':
                    first_type = inner_type.split(',', 1)[0]
                    return resolve(first_type.strip())
                return field_type

            if field_type.startswith('typing.Optional['):
                field_type = resolve(field_type)

            if field_type.startswith('typing.List[') or field_type.startswith('typing.Set['):
                field_type = resolve(field_type)
                array_range = random.randint(1, 5)
                return f"[{', '.join([generate_value(field_type) for _ in range(array_range)])}]"
            elif field_type.startswith('typing.Dict['):
                field_type = resolve(field_type)
                dict_range = random.randint(1, 5)
                dict_data = {}
                for _ in range(dict_range):
                    dict_data[''.join([chr(random.randint(97, 122)) for _ in range(0, 20)])] = generate_value(field_type)
                return f"{{{', '.join([chr(39)+key+chr(39)+f': {value}' for key, value in dict_data.items()])}}}"
            elif field_type.startswith('typing.Union['):
                field_type = resolve(field_type)
                return generate_value(field_type)
            return test_values.get(field_type, 'Test_' + field_type + '.create_instance()')

        return generate_value(field_type)

    def generate_test_class(self, package_name: str, class_name: str, fields: List[Dict[str, str]], 
                           import_types: Set[str]) -> None:
        """Generates a unit test class for a Python dataclass"""
        test_class_name = f"Test_{class_name}"
        tests_package_name = "test_" + package_name.replace('.', '_').lower()
        test_class_definition = process_template(
            "structuretopython/test_class.jinja",
            package_name=package_name,
            class_name=class_name,
            test_class_name=test_class_name,
            fields=fields,
            import_types=import_types,
            avro_annotation=self.avro_annotation,
            dataclasses_json_annotation=self.dataclasses_json_annotation
        )

        base_dir = os.path.join(self.output_dir, "tests")
        test_file_path = os.path.join(base_dir, f"{tests_package_name.replace('.', '_').lower()}.py")
        if not os.path.exists(os.path.dirname(test_file_path)):
            os.makedirs(os.path.dirname(test_file_path), exist_ok=True)
        with open(test_file_path, 'w', encoding='utf-8') as file:
            file.write(test_class_definition)

    def generate_test_enum(self, package_name: str, class_name: str, symbols: List[str]) -> None:
        """Generates a unit test class for a Python enum"""
        test_class_name = f"Test_{class_name}"
        tests_package_name = "test_" + package_name.replace('.', '_').lower()
        test_class_definition = process_template(
            "structuretopython/test_enum.jinja",
            package_name=package_name,
            class_name=class_name,
            test_class_name=test_class_name,
            symbols=symbols
        )
        base_dir = os.path.join(self.output_dir, "tests")
        test_file_path = os.path.join(base_dir, f"{tests_package_name.replace('.', '_').lower()}.py")
        if not os.path.exists(os.path.dirname(test_file_path)):
            os.makedirs(os.path.dirname(test_file_path), exist_ok=True)
        with open(test_file_path, 'w', encoding='utf-8') as file:
            file.write(test_class_definition)

    def write_to_file(self, package: str, class_name: str, python_code: str):
        """Writes a Python class to a file"""
        # Add 'struct' module to the package path
        full_package = f"{package}.struct"
        parent_package_name = '.'.join(full_package.split('.')[:-1])
        parent_package_path = os.sep.join(parent_package_name.split('.')).lower()
        directory_path = os.path.join(self.output_dir, "src", parent_package_path)
        if not os.path.exists(directory_path):
            os.makedirs(directory_path, exist_ok=True)
        file_path = os.path.join(directory_path, f"{class_name.lower()}.py")

        with open(file_path, 'w', encoding='utf-8') as file:
            file.write(python_code)

    def write_init_files(self):
        """Writes __init__.py files to the output directories"""
        def organize_generated_types():
            generated_types_tree = {}
            for generated_type, _ in self.generated_types.items():
                parts = generated_type.split('.')
                if len(parts) < 2:
                    continue
                class_name = parts[-1]
                module_name = parts[-2]
                package_parts = parts[:-2]
                current_node = generated_types_tree
                for part in package_parts:
                    if part not in current_node:
                        current_node[part] = {}
                    current_node = current_node[part]
                current_node[module_name] = class_name
            return generated_types_tree

        def collect_class_names(node):
            class_names = []
            for key, value in node.items():
                if isinstance(value, dict):
                    class_names.extend(collect_class_names(value))
                else:
                    class_names.append(value)
            return class_names

        def write_init_files_recursive(generated_types_tree, current_package: str):
            import_statements = []
            all_statement = []
            for package_or_module_name, content in generated_types_tree.items():
                if isinstance(content, dict):
                    class_names = collect_class_names(content)
                    if class_names:
                        import_statements.append(f"from .{package_or_module_name} import {', '.join(class_names)}")
                        all_statement.extend([f'"{name}"' for name in class_names])
                    write_init_files_recursive(content, current_package + ('.' if current_package else '') + package_or_module_name)
                else:
                    class_name = content
                    import_statements.append(f"from .{package_or_module_name} import {class_name}")
                    all_statement.append(f'"{class_name}"')
            if current_package and (import_statements or all_statement):
                package_path = os.path.join(self.output_dir, 'src', current_package.replace('.', os.sep).lower())
                init_file_path = os.path.join(package_path, '__init__.py')
                if not os.path.exists(package_path):
                    os.makedirs(package_path, exist_ok=True)
                with open(init_file_path, 'w', encoding='utf-8') as file:
                    file.write('\n'.join(import_statements) + '\n\n__all__ = [' + ', '.join(all_statement) + ']\n')

        write_init_files_recursive(organize_generated_types(), '')

    def write_pyproject_toml(self):
        """Writes pyproject.toml file to the output directory"""
        pyproject_content = process_template(
            "structuretopython/pyproject_toml.jinja",
            package_name=self.base_package.replace('_', '-'),
            dataclasses_json_annotation=self.dataclasses_json_annotation,
            avro_annotation=self.avro_annotation
        )
        with open(os.path.join(self.output_dir, 'pyproject.toml'), 'w', encoding='utf-8') as file:
            file.write(pyproject_content)

    def convert_schemas(self, structure_schemas: List, output_dir: str):
        """ Converts JSON Structure schemas to Python dataclasses"""
        self.output_dir = output_dir
        if not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir, exist_ok=True)

        # Register all schema IDs first
        for structure_schema in structure_schemas:
            self.register_schema_ids(structure_schema)

        for structure_schema in structure_schemas:
            self.schema_doc = structure_schema
            if 'definitions' in structure_schema:
                self.definitions = structure_schema['definitions']

            if 'enum' in structure_schema:
                self.generate_enum(structure_schema, structure_schema.get('name', 'Enum'), 
                                 structure_schema.get('namespace', ''), write_file=True)
            elif structure_schema.get('type') == 'object':
                self.generate_class(structure_schema, structure_schema.get('namespace', ''), write_file=True)
            elif structure_schema.get('type') == 'choice':
                self.generate_choice(structure_schema, structure_schema.get('namespace', ''), write_file=True)
            elif structure_schema.get('type') == 'map':
                self.generate_map_alias(structure_schema, structure_schema.get('namespace', ''), write_file=True)

        self.write_init_files()
        self.write_pyproject_toml()

    def convert(self, structure_schema_path: str, output_dir: str):
        """Converts JSON Structure schema to Python dataclasses"""
        with open(structure_schema_path, 'r', encoding='utf-8') as file:
            schema = json.load(file)
        if isinstance(schema, dict):
            schema = [schema]
        return self.convert_schemas(schema, output_dir)


def convert_structure_to_python(structure_schema_path, py_file_path, package_name='', dataclasses_json_annotation=False, avro_annotation=False):
    """Converts JSON Structure schema to Python dataclasses"""
    if not package_name:
        package_name = os.path.splitext(os.path.basename(structure_schema_path))[0].lower().replace('-', '_')

    structure_to_python = StructureToPython(package_name, dataclasses_json_annotation=dataclasses_json_annotation, avro_annotation=avro_annotation)
    structure_to_python.convert(structure_schema_path, py_file_path)


def convert_structure_schema_to_python(structure_schema, py_file_path, package_name='', dataclasses_json_annotation=False):
    """Converts JSON Structure schema to Python dataclasses"""
    structure_to_python = StructureToPython(package_name, dataclasses_json_annotation=dataclasses_json_annotation)
    if isinstance(structure_schema, dict):
        structure_schema = [structure_schema]
    structure_to_python.convert_schemas(structure_schema, py_file_path)



# pylint: disable=too-many-arguments, too-many-locals, too-many-branches, too-many-statements, line-too-long

"""Generates C++ code from JSON Structure schema"""
import json
import os
from typing import Dict, List, Union, Set, Optional, Any, cast

from avrotize.common import pascal, process_template

INDENT = '    '

JsonNode = Dict[str, 'JsonNode'] | List['JsonNode'] | str | None


class StructureToCpp:
    """Converts JSON Structure schema to C++ code, including JSON serialization methods"""

    def __init__(self, base_namespace: str = '') -> None:
        self.base_namespace = base_namespace
        self.output_dir = os.getcwd()
        self.generated_types_namespace: Dict[str, str] = {}
        self.generated_types_cpp_namespace: Dict[str, str] = {}
        self.json_annotation = True  # JSON Structure always has JSON serialization
        self.generated_files: List[str] = []
        self.test_files: List[str] = []
        self.schema_doc: JsonNode = None
        self.schema_registry: Dict[str, Dict] = {}
        self.generated_structure_types: Dict[str, Dict[str, Union[str, Dict, List]]] = {}

    def safe_identifier(self, name: str) -> str:
        """Converts a name to a safe C++ identifier"""
        reserved_words = [
            'alignas', 'alignof', 'and', 'and_eq', 'asm', 'atomic_cancel', 'atomic_commit', 'atomic_noexcept', 'auto',
            'bitand', 'bitor', 'bool', 'break', 'case', 'catch', 'char', 'char8_t', 'char16_t', 'char32_t', 'class',
            'compl', 'concept', 'const', 'consteval', 'constexpr', 'constinit', 'const_cast', 'continue', 'co_await',
            'co_return', 'co_yield', 'decltype', 'default', 'delete', 'do', 'double', 'dynamic_cast', 'else', 'enum',
            'explicit', 'export', 'extern', 'false', 'float', 'for', 'friend', 'goto', 'if', 'inline', 'int', 'long',
            'mutable', 'namespace', 'new', 'noexcept', 'not', 'not_eq', 'nullptr', 'operator', 'or', 'or_eq', 'private',
            'protected', 'public', 'reflexpr', 'register', 'reinterpret_cast', 'requires', 'return', 'short', 'signed',
            'sizeof', 'static', 'static_assert', 'static_cast', 'struct', 'switch', 'synchronized', 'template', 'this',
            'thread_local', 'throw', 'true', 'try', 'typedef', 'typeid', 'typename', 'union', 'unsigned', 'using',
            'virtual', 'void', 'volatile', 'wchar_t', 'while', 'xor', 'xor_eq'
        ]
        if name in reserved_words:
            return f"{name}_"
        return name

    def map_primitive_to_cpp(self, structure_type: str, is_optional: bool) -> str:
        """Maps JSON Structure primitive types to C++ types"""
        optional_mapping = {
            'null': 'std::optional<std::monostate>',
            'boolean': 'std::optional<bool>',
            'string': 'std::optional<std::string>',
            'integer': 'std::optional<int>',
            'number': 'std::optional<double>',
            'int8': 'std::optional<int8_t>',
            'uint8': 'std::optional<uint8_t>',
            'int16': 'std::optional<int16_t>',
            'uint16': 'std::optional<uint16_t>',
            'int32': 'std::optional<int32_t>',
            'uint32': 'std::optional<uint32_t>',
            'int64': 'std::optional<int64_t>',
            'uint64': 'std::optional<uint64_t>',
            'float': 'std::optional<float>',
            'double': 'std::optional<double>',
            'binary': 'std::optional<std::vector<uint8_t>>',
            'date': 'std::optional<std::chrono::system_clock::time_point>',
            'time': 'std::optional<std::chrono::milliseconds>',
            'datetime': 'std::optional<std::chrono::system_clock::time_point>',
            'timestamp': 'std::optional<std::chrono::system_clock::time_point>',
            'duration': 'std::optional<std::chrono::milliseconds>',
            'uuid': 'std::optional<boost::uuids::uuid>',
            'uri': 'std::optional<std::string>',
            'jsonpointer': 'std::optional<std::string>',
            'decimal': 'std::optional<std::string>',
            'any': 'nlohmann::json'
        }
        required_mapping = {
            'null': 'std::monostate',
            'boolean': 'bool',
            'string': 'std::string',
            'integer': 'int',
            'number': 'double',
            'int8': 'int8_t',
            'uint8': 'uint8_t',
            'int16': 'int16_t',
            'uint16': 'uint16_t',
            'int32': 'int32_t',
            'uint32': 'uint32_t',
            'int64': 'int64_t',
            'uint64': 'uint64_t',
            'float': 'float',
            'double': 'double',
            'binary': 'std::vector<uint8_t>',
            'date': 'std::chrono::system_clock::time_point',
            'time': 'std::chrono::milliseconds',
            'datetime': 'std::chrono::system_clock::time_point',
            'timestamp': 'std::chrono::system_clock::time_point',
            'duration': 'std::chrono::milliseconds',
            'uuid': 'boost::uuids::uuid',
            'uri': 'std::string',
            'jsonpointer': 'std::string',
            'decimal': 'std::string',
            'any': 'nlohmann::json'
        }
        if '.' in structure_type:
            type_name = structure_type.split('.')[-1]
            package_name = '::'.join(structure_type.split('.')[:-1]).lower()
            structure_type = self.get_qualified_name(package_name, type_name)
        if structure_type in self.generated_types_namespace:
            return self.get_qualified_name(self.base_namespace, structure_type)
        else:
            return required_mapping.get(structure_type, structure_type) if not is_optional else optional_mapping.get(structure_type, f'std::optional<{required_mapping.get(structure_type, structure_type)}>')

    def get_qualified_name(self, namespace: str, name: str) -> str:
        """Concatenates namespace and name using a double colon separator"""
        return f"{namespace}::{name}" if namespace else name
    
    def concat_namespace(self, namespace: str, name: str) -> str:
        """Concatenates namespace and name using a double colon separator"""
        if namespace and name:
            return f"{namespace}::{name}"
        elif namespace:
            return namespace
        return name

    def resolve_ref(self, ref: str, context_schema: Optional[Dict] = None) -> Optional[Dict]:
        """ Resolves a $ref to the actual schema definition """
        # Check if it's an absolute URI reference (schema with $id)
        if not ref.startswith('#/'):
            # Try to resolve from schema registry
            if ref in self.schema_registry:
                return self.schema_registry[ref]
            return None
        
        # Handle fragment-only references (internal to document)
        path = ref[2:].split('/')
        schema = context_schema if context_schema else self.schema_doc
        
        for part in path:
            if not isinstance(schema, dict) or part not in schema:
                return None
            schema = schema[part]
        
        return schema

    def register_schema_ids(self, schema: Dict, base_uri: str = '') -> None:
        """ Recursively registers schemas with $id keywords """
        if not isinstance(schema, dict):
            return
        
        # Register this schema if it has an $id
        if '$id' in schema:
            schema_id = schema['$id']
            # Handle relative URIs
            if base_uri and not schema_id.startswith(('http://', 'https://', 'urn:')):
                from urllib.parse import urljoin
                schema_id = urljoin(base_uri, schema_id)
            self.schema_registry[schema_id] = schema
            base_uri = schema_id  # Update base URI for nested schemas
        
        # Recursively process definitions
        if 'definitions' in schema:
            for def_name, def_schema in schema['definitions'].items():
                if isinstance(def_schema, dict):
                    self.register_schema_ids(def_schema, base_uri)
        
        # Recursively process properties
        if 'properties' in schema:
            for prop_name, prop_schema in schema['properties'].items():
                if isinstance(prop_schema, dict):
                    self.register_schema_ids(prop_schema, base_uri)
        
        # Recursively process items, values, etc.
        for key in ['items', 'values', 'additionalProperties']:
            if key in schema and isinstance(schema[key], dict):
                self.register_schema_ids(schema[key], base_uri)

    def convert_structure_type_to_cpp(self, class_name: str, field_name: str, structure_type: JsonNode, parent_namespace: str, nullable: bool = False) -> str:
        """Converts JSON Structure type to C++ type"""
        if isinstance(structure_type, str):
            return self.map_primitive_to_cpp(structure_type, nullable)
        elif isinstance(structure_type, list):
            # Handle type unions
            non_null_types = [t for t in structure_type if t != 'null']
            is_nullable = 'null' in structure_type
            if len(non_null_types) == 1:
                if isinstance(non_null_types[0], str):
                    return self.map_primitive_to_cpp(non_null_types[0], is_nullable)
                else:
                    cpp_type = self.convert_structure_type_to_cpp(class_name, field_name, non_null_types[0], parent_namespace, nullable=False)
                    if is_nullable:
                        return f'std::optional<{cpp_type}>'
                    return cpp_type
            else:
                types: List[str] = [self.convert_structure_type_to_cpp(class_name, field_name, t, parent_namespace, nullable=False) for t in non_null_types]
                return 'std::variant<' + ', '.join(types) + '>'
        elif isinstance(structure_type, dict):
            # Handle $ref
            if '$ref' in structure_type:
                ref_schema = self.resolve_ref(structure_type['$ref'], self.schema_doc if isinstance(self.schema_doc, dict) else None)
                if ref_schema:
                    ref_path = structure_type['$ref'].split('/')
                    type_name = ref_path[-1]
                    ref_namespace = '.'.join(ref_path[2:-1]) if len(ref_path) > 3 else parent_namespace
                    return self.generate_class_or_choice(ref_schema, ref_namespace, write_file=True, explicit_name=type_name)
                return 'nlohmann::json'
            
            # Handle enum keyword
            if 'enum' in structure_type:
                return self.generate_enum(structure_type, field_name, parent_namespace, write_file=True)
            
            # Handle type keyword
            if 'type' not in structure_type:
                return 'nlohmann::json'
            
            struct_type = structure_type['type']
            
            if struct_type == 'object':
                return self.generate_class(structure_type, parent_namespace, write_file=True)
            elif struct_type == 'array':
                items_type = self.convert_structure_type_to_cpp(class_name, field_name+'Item', structure_type.get('items', {'type': 'any'}), parent_namespace, nullable=False)
                return f"std::vector<{items_type}>"
            elif struct_type == 'set':
                items_type = self.convert_structure_type_to_cpp(class_name, field_name+'Item', structure_type.get('items', {'type': 'any'}), parent_namespace, nullable=False)
                return f"std::set<{items_type}>"
            elif struct_type == 'map':
                values_type = self.convert_structure_type_to_cpp(class_name, field_name+'Value', structure_type.get('values', {'type': 'any'}), parent_namespace, nullable=False)
                return f"std::map<std::string, {values_type}>"
            elif struct_type == 'choice':
                return self.generate_choice(structure_type, parent_namespace, write_file=True)
            elif struct_type == 'tuple':
                return self.generate_tuple(structure_type, parent_namespace, write_file=True)
            else:
                return self.convert_structure_type_to_cpp(class_name, field_name, struct_type, parent_namespace, nullable)
        return 'nlohmann::json'

    def generate_class_or_choice(self, structure_schema: Dict, parent_namespace: str, write_file: bool = True, explicit_name: str = '') -> str:
        """Generates a Class or Choice"""
        struct_type = structure_schema.get('type', 'object')
        if struct_type == 'object':
            return self.generate_class(structure_schema, parent_namespace, write_file, explicit_name=explicit_name)
        elif struct_type == 'choice':
            return self.generate_choice(structure_schema, parent_namespace, write_file, explicit_name=explicit_name)
        elif struct_type == 'tuple':
            return self.generate_tuple(structure_schema, parent_namespace, write_file, explicit_name=explicit_name)
        elif struct_type in ('map', 'array', 'set'):
            # For root-level container types, generate a type alias
            return self.generate_container_alias(structure_schema, parent_namespace, write_file, explicit_name=explicit_name)
        return 'nlohmann::json'

    def generate_class(self, structure_schema: Dict, parent_namespace: str, write_file: bool, explicit_name: str = '') -> str:
        """Generates a C++ class from a JSON Structure object type"""
        class_definition = ''
        
        # Get name and namespace
        class_name = pascal(explicit_name if explicit_name else structure_schema.get('name', 'UnnamedClass'))
        schema_namespace = structure_schema.get('namespace', parent_namespace)
        namespace = self.concat_namespace(self.base_namespace, schema_namespace.replace('.', '::'))
        
        qualified_class_name = self.get_qualified_name(namespace, class_name)
        if qualified_class_name in self.generated_types_namespace:
            return qualified_class_name
        
        self.generated_types_namespace[qualified_class_name] = schema_namespace
        self.generated_types_cpp_namespace[qualified_class_name] = "class"
        self.generated_structure_types[qualified_class_name] = structure_schema

        # Track the includes for member types
        member_includes = set()

        # Generate class documentation
        if 'description' in structure_schema or 'doc' in structure_schema:
            doc = structure_schema.get('description', structure_schema.get('doc', ''))
            class_definition += f"// {doc}\n"
        
        # Check if this is an abstract type
        is_abstract = structure_schema.get('abstract', False)
        
        class_definition += f"class {class_name} {{\n"
        class_definition += "public:\n"
        
        # Generate properties
        properties = structure_schema.get('properties', {})
        required_props = structure_schema.get('required', [])
        
        for prop_name, prop_schema in properties.items():
            field_name = self.safe_identifier(prop_name)
            is_required = prop_name in required_props if not isinstance(required_props, list) or len(required_props) == 0 or not isinstance(required_props[0], list) else any(prop_name in req_set for req_set in required_props)
            
            # Convert to C++ type
            field_type = self.convert_structure_type_to_cpp(class_name, field_name, prop_schema, schema_namespace, nullable=not is_required)
            
            # Add documentation
            if 'description' in prop_schema or 'doc' in prop_schema:
                field_doc = prop_schema.get('description', prop_schema.get('doc', ''))
                class_definition += f"{INDENT}// {field_doc}\n"
            
            class_definition += f"{INDENT}{field_type} {field_name};\n"
        
        # Add default constructor
        class_definition += f"\npublic:\n"
        class_definition += f"{INDENT}{class_name}() = default;\n"

        # Add JSON serialization methods
        class_definition += process_template("structuretocpp/dataclass_body.jinja", 
                                            class_name=class_name, 
                                            json_annotation=self.json_annotation)
        
        if self.json_annotation:
            class_definition += self.generate_to_json_method(class_name)
        
        class_definition += "};\n\n"

        # Create includes
        includes = self.generate_includes(member_includes)

        if write_file:
            self.write_to_file(namespace, class_name, includes, class_definition)
            self.generate_unit_test(class_name, properties, namespace, required_props)
        
        return qualified_class_name

    def generate_enum(self, structure_schema: Dict, field_name: str, parent_namespace: str, write_file: bool) -> str:
        """Generates a C++ enum from a JSON Structure enum"""
        enum_definition = ''
        
        # Generate enum name
        enum_name = pascal(structure_schema.get('name', field_name + 'Enum'))
        schema_namespace = structure_schema.get('namespace', parent_namespace)
        namespace = self.concat_namespace(self.base_namespace, schema_namespace.replace('.', '::'))
        
        qualified_enum_name = self.get_qualified_name(namespace, enum_name)
        if qualified_enum_name in self.generated_types_namespace:
            return qualified_enum_name
        
        self.generated_types_namespace[qualified_enum_name] = schema_namespace
        self.generated_types_cpp_namespace[qualified_enum_name] = "enum"
        
        if 'description' in structure_schema or 'doc' in structure_schema:
            doc = structure_schema.get('description', structure_schema.get('doc', ''))
            enum_definition += f"// {doc}\n"
        
        symbols = structure_schema.get('enum', [])
        enum_definition += f"enum class {enum_name} {{\n"
        for symbol in symbols:
            enum_definition += f"{INDENT}{self.safe_identifier(str(symbol))},\n"
        enum_definition += "};\n\n"
        
        if write_file:
            self.write_to_file(namespace, enum_name, "", enum_definition)
        
        return qualified_enum_name

    def generate_choice(self, structure_schema: Dict, parent_namespace: str, write_file: bool, explicit_name: str = '') -> str:
        """Generates a choice (discriminated union) type"""
        # For now, return variant of the choice types
        choices = structure_schema.get('choices', {})
        choice_types = []
        
        for choice_name, choice_schema in choices.items():
            if isinstance(choice_schema, dict):
                if '$ref' in choice_schema:
                    ref_schema = self.resolve_ref(choice_schema['$ref'], self.schema_doc if isinstance(self.schema_doc, dict) else None)
                    if ref_schema:
                        ref_path = choice_schema['$ref'].split('/')
                        type_name = ref_path[-1]
                        ref_namespace = '.'.join(ref_path[2:-1]) if len(ref_path) > 3 else parent_namespace
                        qualified_name = self.generate_class(ref_schema, ref_namespace, write_file=True, explicit_name=type_name)
                        choice_types.append(qualified_name)
                elif 'type' in choice_schema:
                    choice_type = self.convert_structure_type_to_cpp('Choice', choice_name, choice_schema, parent_namespace, nullable=False)
                    choice_types.append(choice_type)
        
        if len(choice_types) == 0:
            return 'nlohmann::json'
        elif len(choice_types) == 1:
            return choice_types[0]
        else:
            return f"std::variant<{', '.join(choice_types)}>"

    def generate_tuple(self, structure_schema: Dict, parent_namespace: str, write_file: bool, explicit_name: str = '') -> str:
        """Generates a tuple type"""
        # Tuples serialize as JSON arrays
        # For now, use std::tuple
        properties = structure_schema.get('properties', {})
        tuple_order = structure_schema.get('tuple', [])
        
        tuple_types = []
        for prop_name in tuple_order:
            if prop_name in properties:
                prop_schema = properties[prop_name]
                prop_type = self.convert_structure_type_to_cpp('Tuple', prop_name, prop_schema, parent_namespace, nullable=False)
                tuple_types.append(prop_type)
        
        if len(tuple_types) == 0:
            return 'nlohmann::json'
        
        return f"std::tuple<{', '.join(tuple_types)}>"

    def generate_container_alias(self, structure_schema: Dict, parent_namespace: str, write_file: bool, explicit_name: str = '') -> str:
        """Generates a type alias for root-level container types (map, array, set)"""
        struct_type = structure_schema.get('type', 'map')
        
        # Get name and namespace
        class_name = pascal(explicit_name if explicit_name else structure_schema.get('name', f'Root{struct_type.capitalize()}'))
        schema_namespace = structure_schema.get('namespace', parent_namespace)
        namespace = self.concat_namespace(self.base_namespace, schema_namespace.replace('.', '::'))
        
        qualified_name = self.get_qualified_name(namespace, class_name)
        if qualified_name in self.generated_types_namespace:
            return qualified_name
        
        self.generated_types_namespace[qualified_name] = schema_namespace
        self.generated_types_cpp_namespace[qualified_name] = "alias"
        self.generated_structure_types[qualified_name] = structure_schema
        
        # Determine the underlying type
        if struct_type == 'map':
            values_type = self.convert_structure_type_to_cpp(class_name, 'Value', structure_schema.get('values', {'type': 'any'}), schema_namespace, nullable=False)
            underlying_type = f"std::map<std::string, {values_type}>"
        elif struct_type == 'array':
            items_type = self.convert_structure_type_to_cpp(class_name, 'Item', structure_schema.get('items', {'type': 'any'}), schema_namespace, nullable=False)
            underlying_type = f"std::vector<{items_type}>"
        elif struct_type == 'set':
            items_type = self.convert_structure_type_to_cpp(class_name, 'Item', structure_schema.get('items', {'type': 'any'}), schema_namespace, nullable=False)
            underlying_type = f"std::set<{items_type}>"
        else:
            underlying_type = 'nlohmann::json'
        
        # Generate type alias
        alias_definition = ''
        if 'description' in structure_schema or 'doc' in structure_schema:
            doc = structure_schema.get('description', structure_schema.get('doc', ''))
            alias_definition += f"// {doc}\n"
        
        alias_definition += f"using {class_name} = {underlying_type};\n\n"
        
        if write_file:
            self.write_to_file(namespace, class_name, "", alias_definition)
        
        return qualified_name

    def generate_to_json_method(self, class_name: str) -> str:
        """Generates the to_json method for the class"""
        method_definition = f"\nnlohmann::json to_json() const {{\n"
        method_definition += f"{INDENT}return nlohmann::json(*this);\n"
        method_definition += f"}}\n"
        return method_definition

    def generate_includes(self, member_includes: set) -> str:
        """Generates the include statements for the member types"""
        includes = '\n'.join([f'#include "{include.replace("::", "/")}.hpp"' for include in member_includes])
        return includes
    
    def generate_unit_test(self, class_name: str, properties: Dict, namespace: str, required_props: List) -> None:
        """Generates a unit test for a given class"""
        test_definition = f'#include <gtest/gtest.h>\n'
        test_definition += f'#include "{namespace.replace("::", "/")}/{class_name}.hpp"\n\n'
        test_definition += f'TEST({class_name}Test, PropertiesTest) {{\n'
        test_definition += f'{INDENT}{namespace}::{class_name} instance;\n'

        for prop_name, prop_schema in properties.items():
            field_name = self.safe_identifier(prop_name)
            is_required = prop_name in required_props if not isinstance(required_props, list) or len(required_props) == 0 or not isinstance(required_props[0], list) else any(prop_name in req_set for req_set in required_props)
            test_value = self.get_test_value(prop_schema, is_required)
            test_definition += f'{INDENT}instance.{field_name} = {test_value};\n'
            test_definition += f'{INDENT}EXPECT_EQ(instance.{field_name}, {test_value});\n'

        test_definition += '}\n'

        test_dir = os.path.join(self.output_dir, "tests")
        if not os.path.exists(test_dir):
            os.makedirs(test_dir, exist_ok=True)

        test_file_path = os.path.join(test_dir, f"{class_name}_test.cpp")
        with open(test_file_path, 'w', encoding='utf-8') as file:
            file.write(test_definition)

        self.test_files.append(test_file_path.replace(os.sep, '/'))

    def get_test_value(self, structure_schema: JsonNode, is_required: bool) -> str:
        """Returns a default test value based on the Structure type"""
        if isinstance(structure_schema, str):
            test_values = {
                'string': '"test_string"',
                'boolean': 'true',
                'integer': '42',
                'number': '3.14',
                'int8': '42',
                'uint8': '42',
                'int16': '42',
                'uint16': '42',
                'int32': '42',
                'uint32': '42',
                'int64': '42LL',
                'uint64': '42ULL',
                'float': '3.14f',
                'double': '3.14',
                'binary': 'std::vector<uint8_t>{0x01, 0x02, 0x03}',
                'date': 'std::chrono::system_clock::now()',
                'time': 'std::chrono::milliseconds(123456)',
                'datetime': 'std::chrono::system_clock::now()',
                'timestamp': 'std::chrono::system_clock::now()',
                'duration': 'std::chrono::milliseconds(1000)',
                'uuid': 'boost::uuids::random_generator()()',
                'uri': '"https://example.com"',
                'jsonpointer': '"/path/to/field"',
                'decimal': '"123.45"',
                'any': 'nlohmann::json::object()',
                'null': 'std::monostate()',
            }
            return test_values.get(structure_schema, '"test"')

        elif isinstance(structure_schema, list):
            # For unions, use the first non-null type
            non_null_types = [t for t in structure_schema if t != 'null']
            if non_null_types:
                return self.get_test_value(non_null_types[0], True)
            return 'std::monostate()'

        elif isinstance(structure_schema, dict):
            struct_type = structure_schema.get('type', 'any')
            
            if 'enum' in structure_schema:
                symbols = structure_schema.get('enum', [])
                if symbols:
                    return f"{pascal(structure_schema.get('name', 'Enum'))}_::{self.safe_identifier(str(symbols[0]))}"
                return '"test"'
            
            if struct_type == 'object':
                return f"{pascal(structure_schema.get('name', 'Object'))}()"
            elif struct_type == 'array':
                item_type = self.convert_structure_type_to_cpp('Test', 'Item', structure_schema.get('items', {'type': 'any'}), '', False)
                return f"std::vector<{item_type}>{{}}"
            elif struct_type == 'set':
                item_type = self.convert_structure_type_to_cpp('Test', 'Item', structure_schema.get('items', {'type': 'any'}), '', False)
                return f"std::set<{item_type}>{{}}"
            elif struct_type == 'map':
                value_type = self.convert_structure_type_to_cpp('Test', 'Value', structure_schema.get('values', {'type': 'any'}), '', False)
                return f"std::map<std::string, {value_type}>{{}}"
            elif struct_type == 'tuple':
                return '"tuple_test"'
            else:
                return self.get_test_value(struct_type, is_required)

        return '"test"'

    def write_to_file(self, namespace: str, name: str, includes: str, definition: str) -> None:
        """Writes a C++ class or enum to a file"""
        directory_path = os.path.join(
            self.output_dir, "include", namespace.replace('::', os.sep))
        if not os.path.exists(directory_path):
            os.makedirs(directory_path, exist_ok=True)
        file_path = os.path.join(directory_path, f"{name}.hpp")

        with open(file_path, 'w', encoding='utf-8') as file:
            file.write("#pragma once\n")
            if self.json_annotation:
                file.write("#include <nlohmann/json.hpp>\n")
            if "std::vector" in definition:
                file.write("#include <vector>\n")
            if "std::set" in definition:
                file.write("#include <set>\n")
            if "std::map" in definition:
                file.write("#include <map>\n")
            if "std::tuple" in definition:
                file.write("#include <tuple>\n")
            if "std::variant" in definition:
                file.write("#include <variant>\n")
            if "std::optional" in definition:
                file.write("#include <optional>\n")
            file.write("#include <stdexcept>\n")
            if "std::chrono" in definition:
                file.write("#include <chrono>\n")
            if "boost::uuid" in definition:
                file.write("#include <boost/uuid/uuid.hpp>\n")
                file.write("#include <boost/uuid/uuid_io.hpp>\n")
                file.write("#include <boost/uuid/uuid_generators.hpp>\n")
            if includes:
                file.write(includes + '\n')
            if namespace:
                file.write(f"\nnamespace {namespace} {{\n\n")
            file.write(definition)
            if namespace:
                file.write(f"}} // namespace {namespace}\n")

        # Collect the generated file names
        self.generated_files.append(file_path.replace(os.sep, '/'))

    def generate_cmake_lists(self, project_name: str) -> None:
        """Generates a CMakeLists.txt file"""
        cmake_content = process_template("structuretocpp/CMakeLists.txt.jinja", 
                                        project_name=project_name, 
                                        json_annotation=self.json_annotation)
        cmake_path = os.path.join(self.output_dir, 'CMakeLists.txt')
        with open(cmake_path, 'w', encoding='utf-8') as file:
            file.write(cmake_content)
    
    def generate_vcpkg_json(self) -> None:
        """Generates a vcpkg.json file"""
        vcpkg_json = process_template("structuretocpp/vcpkg.json.jinja", 
                                      project_name=self.base_namespace, 
                                      json_annotation=self.json_annotation)
        vcpkg_json_path = os.path.join(self.output_dir, 'vcpkg.json')
        with open(vcpkg_json_path, 'w', encoding='utf-8') as file:
            file.write(vcpkg_json)
            
    def generate_build_scripts(self) -> None:
        """Generates build scripts for Windows and Linux"""
        build_script_linux = process_template("structuretocpp/build.sh.jinja")
        build_script_windows = process_template("structuretocpp/build.bat.jinja")        
        script_path_linux = os.path.join(self.output_dir, 'build.sh')
        script_path_windows = os.path.join(self.output_dir, 'build.bat')

        with open(script_path_linux, 'w', encoding='utf-8') as file:
            file.write(build_script_linux)

        with open(script_path_windows, 'w', encoding='utf-8') as file:
            file.write(build_script_windows)

    def convert_schema(self, schema: Union[Dict, List], output_dir: str) -> None:
        """Converts JSON Structure schema to C++"""
        if not isinstance(schema, list):
            schema = [schema]
        if not os.path.exists(output_dir):
            os.makedirs(output_dir, exist_ok=True)
        self.output_dir = output_dir
        
        # Register all schema IDs first
        for structure_schema in (s for s in schema if isinstance(s, dict)):
            self.register_schema_ids(structure_schema)
        
        # Process each schema
        for structure_schema in (s for s in schema if isinstance(s, dict)):
            self.schema_doc = structure_schema
            
            # Process root type
            if 'type' in structure_schema:
                self.generate_class_or_choice(structure_schema, '', write_file=True)
            elif '$root' in structure_schema:
                root_ref = structure_schema['$root']
                root_schema = self.resolve_ref(root_ref, structure_schema)
                if root_schema:
                    ref_path = root_ref.split('/')
                    type_name = ref_path[-1]
                    ref_namespace = '.'.join(ref_path[2:-1]) if len(ref_path) > 3 else ''
                    self.generate_class_or_choice(root_schema, ref_namespace, write_file=True, explicit_name=type_name)
            
            # Process definitions
            if 'definitions' in structure_schema:
                self.process_definitions(structure_schema['definitions'], '')
        
        self.generate_cmake_lists(self.base_namespace)
        self.generate_build_scripts()
        self.generate_vcpkg_json()

    def process_definitions(self, definitions: Dict, namespace_path: str) -> None:
        """ Processes the definitions section recursively """
        for name, definition in definitions.items():
            if isinstance(definition, dict):
                if 'type' in definition or 'enum' in definition:
                    # This is a type definition
                    current_namespace = self.concat_namespace(namespace_path, '')
                    # Check if this type was already generated
                    check_namespace = self.concat_namespace(self.base_namespace, current_namespace).replace('.', '::')
                    check_name = pascal(name)
                    check_ref = self.get_qualified_name(check_namespace, check_name)
                    if check_ref not in self.generated_types_namespace:
                        self.generate_class_or_choice(definition, current_namespace, write_file=True, explicit_name=name)
                else:
                    # This is a namespace
                    new_namespace = self.concat_namespace(namespace_path, name)
                    self.process_definitions(definition, new_namespace)

    def convert(self, structure_schema_path: str, output_dir: str) -> None:
        """Converts JSON Structure schema to C++"""
        with open(structure_schema_path, 'r', encoding='utf-8') as file:
            schema = json.load(file)
        self.convert_schema(schema, output_dir)


def convert_structure_to_cpp(structure_schema_path: str, output_dir: str, namespace: str = '', json_annotation: bool = True) -> None:
    """Converts JSON Structure schema to C++ classes"""
    
    if not namespace:
        namespace = os.path.splitext(os.path.basename(structure_schema_path))[0].replace('-', '_')
    
    structure_to_cpp = StructureToCpp(namespace)
    structure_to_cpp.json_annotation = json_annotation
    structure_to_cpp.convert(structure_schema_path, output_dir)

def convert_structure_schema_to_cpp(structure_schema: Dict, output_dir: str, namespace: str = '', json_annotation: bool = True) -> None:
    """Converts JSON Structure schema to C++ classes"""
    structure_to_cpp = StructureToCpp(namespace)
    structure_to_cpp.json_annotation = json_annotation
    structure_to_cpp.convert_schema(structure_schema, output_dir)



from functools import reduce
import json
from typing import Dict, List
import xml.etree.ElementTree as ET
from xml.etree.ElementTree import Element, SubElement, tostring
from xml.dom import minidom

from avrotize.common import is_generic_avro_type

class AvroToXSD:
    def __init__(self, target_namespace: str = ''):
        self.xmlns = {"xs": "http://www.w3.org/2001/XMLSchema"}
        self.union_types: Dict[str, str] = {}
        self.known_types: List[str] = []
        self.common_namespace = ''
        self.target_namespace = target_namespace

    def find_common_namespace(self, namespaces: List[str]) -> str:
        """
        Find the common namespace prefix from a list of namespaces.
        """
        if not namespaces:
            return ''

        def common_prefix(a, b):
            prefix = ''
            for a_char, b_char in zip(a.split('.'), b.split('.')):
                if a_char == b_char:
                    prefix += a_char + '.'
                else:
                    break
            return prefix.rstrip('.')

        return reduce(common_prefix, namespaces)

    def update_common_namespace(self, namespace: str) -> None:
        """
        Update the common namespace based on the provided namespace.
        """
        if not self.common_namespace:
            self.common_namespace = namespace
        else:
            self.common_namespace = self.find_common_namespace([self.common_namespace, namespace])

    def convert_avro_primitive(self, avro_type: str | dict) -> str:
        """Map Avro primitive types to XML schema (XSD) data types."""
        
        if isinstance(avro_type, dict) and 'logicalType' in avro_type:
            type = avro_type['type']
            logical_type = avro_type.get('logicalType')
            
            if logical_type == 'decimal':
                return f"decimal"
            if logical_type == 'timestamp-millis':
                return f"dateTime"
            if logical_type == 'date':
                return f"date"
            if logical_type in {'time-millis', 'time-micros'}:
                return f"time"
            if logical_type == 'uuid':
                return f"string"
        elif isinstance(avro_type, str):
            mapping = {
                'null': 'string',  # Defaulting to string for nullables
                'boolean': 'boolean',
                'int': 'integer',
                'long': 'long',
                'float': 'float',
                'double': 'double',
                'bytes': 'hexBinary',
                'string': 'string',
            }
            
            type = mapping.get(avro_type, '')  # Fallback to string
            if type:
                return f"xs:{type}"
            else:
                return avro_type.split('.')[-1]
        return f"xs:string"
    
    def is_avro_primitive(self, avro_type: str) -> bool:
        """Check if the Avro type is a primitive type."""
        if isinstance(avro_type, dict) and 'logicalType' in avro_type and 'type' in avro_type: 
            return avro_type['type'] in {'int', 'long', 'float', 'double', 'bytes', 'string'}
        elif isinstance(avro_type, str):
            return avro_type in {'null', 'boolean', 'int', 'long', 'float', 'double', 'bytes', 'string'}
        else:
            return False

    def create_element(self, parent: Element, tag: str, **attributes) -> Element:
        """Create an XML element with the proper namespace."""
        return SubElement(parent, f"{{{self.xmlns['xs']}}}{tag}", **attributes)
    
    def create_complex_type(self, parent: Element, **attributes) -> Element:
        """Create an XML complexType element."""
        return self.create_element(parent, "complexType", **attributes)

    def create_fixed(self, schema_root, field_type):
        """ handle Avro 'fixed' type"""
        simple_type = self.create_element(schema_root, "simpleType", name=field_type['name'])
        restriction = self.create_element(simple_type, "restriction", base="xs:hexBinary")
        restriction.set("fixed", "true")
        restriction.set("value", field_type['size'])        

    def create_map(self, schema_root: ET.Element, record_name: str, parent: ET.Element, map_schema: dict):
        """ handle Avro 'map' type"""
        complex_type = self.create_element(parent, "complexType")
        sequence = self.create_element(complex_type, "sequence")
        item = self.create_element(sequence, "element", name="item", minOccurs="0", maxOccurs="unbounded")
        inner_complex_type = self.create_element(item, "complexType")
        inner_sequence = self.create_element(inner_complex_type, "sequence")
        self.create_element(inner_sequence, "element", name="key", type="xs:string")
        map_values = map_schema['values']
        if isinstance(map_values, list):
            item_value = self.create_union(schema_root, record_name, inner_sequence, "value", map_values)
        else:
            item_value = self.create_element(inner_sequence, "element", name="value")
            self.set_field_type(schema_root, record_name, item_value, map_schema['values'])

    def create_union(self, schema_root: ET.Element, record_name: str, parent: ET.Element, field_name: str, field_type: list, insert_annotation: lambda e: None | None = None) -> ET.Element:
        """Create an XML element for union types."""

        def create_or_get_union_simple_type(self, schema_root: ET.Element, parent:Element, types: List[str], **attributes) -> str:
            """Create an XML simpleType element for union types."""

            type_key_list = types.copy()
            type_key_list.sort()
            type_key = ''.join(type_key_list)
            if type_key in self.union_types:
                return self.union_types[type_key]

            name = "And".join([t.capitalize() for t in types])
            simple_type = self.create_element(schema_root, "simpleType", **attributes)
            simple_type.set("name", name)
            union = self.create_element(simple_type, "union")
            union.set("memberTypes", ' '.join([self.convert_avro_primitive(t) for t in types if t != 'null']))
            self.union_types[type_key] = name
            return name

        if isinstance(field_type, list) and is_generic_avro_type(field_type):
            element = self.create_element(parent, "any", minOccurs="0", maxOccurs="unbounded")
            if insert_annotation: 
                insert_annotation(element)
            return element

        non_null_types = [t for t in field_type if t != 'null']
        if len(non_null_types) == 1:
            element = self.create_element(parent, "element", name=field_name)
            if insert_annotation: 
                insert_annotation(element)
            self.set_field_type(schema_root, record_name, element, non_null_types[0])
            return element
        else:
            element = self.create_element(parent, "element", name=field_name)
            if insert_annotation: 
                insert_annotation(element)
            primitives = [t for t in non_null_types if self.is_avro_primitive(t)]
            for primitive in primitives:
                non_null_types.remove(primitive)
            union_type_ref = ''
            if len(primitives) > 0:
                union_type_ref = create_or_get_union_simple_type(self, schema_root, parent, primitives)
                if len(non_null_types) == 0:
                    element.set('type', union_type_ref)
            if len(non_null_types) > 0:
                abstract_complex_type_name = record_name+field_name.capitalize()
                element.set('type', abstract_complex_type_name)
                if not abstract_complex_type_name in self.known_types:
                    self.known_types.append(abstract_complex_type_name)
                    self.create_element(schema_root, "complexType", name=abstract_complex_type_name, abstract="true")
                    if union_type_ref:
                        complex_content_option = self.create_element(schema_root, "complexType", name=abstract_complex_type_name+'1')
                        complex_content = self.create_element(complex_content_option, "complexContent")
                        complex_extension = self.create_element(complex_content, "extension", base=abstract_complex_type_name)
                        complex_sequence = self.create_element(complex_extension, "sequence")
                        complex_element = self.create_element(complex_sequence, "element", name='value', type=union_type_ref)
                    for i, union_type in enumerate(non_null_types):
                        complex_content_option = self.create_element(schema_root, "complexType", name=abstract_complex_type_name+str(i+2))
                        complex_content = self.create_element(complex_content_option, "complexContent")
                        complex_extension = self.create_element(complex_content, "extension", base=abstract_complex_type_name)
                        complex_sequence = self.create_element(complex_extension, "sequence")
                        complex_element = self.create_element(complex_sequence, "element", name=field_name)
                        self.set_field_type(schema_root, record_name, complex_element, union_type)
            return element  

    def create_array(self, schema_root: ET.Element, record_name: str, parent: ET.Element, array_schema: dict):
        """ handle Avro 'array' type """
        complex_type = self.create_element(parent, "complexType")
        sequence = self.create_element(complex_type, "sequence")
        item_type = array_schema['items']
        if isinstance(item_type, list):
            item = self.create_union(schema_root, record_name, sequence, "item", item_type)
            item.set('minOccurs', '0')
            item.set('maxOccurs', 'unbounded')
        else:
            item = self.create_element(sequence, "element", name="item", minOccurs="0", maxOccurs="unbounded")  
            self.set_field_type(schema_root, record_name, item, item_type)

    def create_enum(self, schema_root: ET.Element, enum_schema: dict) -> str:
        """Convert an Avro enum to an XML simpleType."""
        name = enum_schema['name']
        doc = enum_schema.get('doc', '')
        if name in self.known_types:
            return name
        simple_type = self.create_element(schema_root, "simpleType")
        if doc:
            annotation = self.create_element(simple_type, "annotation")
            documentation = self.create_element(annotation, "documentation")
            documentation.text = doc
        simple_type.set('name', name)
        restriction = self.create_element(simple_type, "restriction", base="xs:string")
        for enum_symbol in enum_schema['symbols']:
            self.create_element(restriction, "enumeration", value=enum_symbol)
        self.known_types.append(name)
        return name
    
    def set_field_type(self, schema_root: ET.Element, record_name: str, element: ET.Element, field_type: dict|str):
        """ set the type or create a subtype on the element for the given avro field type"""
        if isinstance(field_type, dict):
            if 'type' in field_type:
                if field_type['type'] == 'record':
                    if 'namespace' in field_type:
                        self.update_common_namespace(field_type['namespace'])
                    type = self.create_record(schema_root, field_type)
                    element.set('type', type)
                elif field_type['type'] == 'enum':
                    if 'namespace' in field_type:
                        self.update_common_namespace(field_type['namespace'])
                    type = self.create_enum(schema_root, field_type)
                    element.set('type', type)
                elif field_type['type'] == 'array':
                    self.create_array(schema_root, record_name, element, field_type)
                elif field_type['type'] == 'map':
                    self.create_map(schema_root, record_name, element, field_type)
                elif field_type['type'] == 'fixed':
                    self.create_fixed(schema_root, field_type)
                else:
                    return self.set_field_type(schema_root, record_name, element, field_type['type'])
            else:
                raise ValueError(f"Invalid field type")
        else:
            element.set('type', self.convert_avro_primitive(field_type))       

    def create_field(self, schema_root: Element, record_name: str, parent: Element, field: dict, attributes_parent: Element) -> ET.Element:
        """Convert an Avro field to an XML element."""
        field_name = field['name']
        field_type = field['type']
        field_doc = field.get('doc', '')
        xmlkind = field.get('xmlkind', 'element')
        if isinstance(field_type,list):
            def ia(e) -> None:
                if field_doc:
                    annotation = self.create_element(e, "annotation")
                    documentation = self.create_element(annotation, "documentation")
                    documentation.text = field_doc    
            element = self.create_union(schema_root, record_name, parent, field_name, field_type, ia)
        else:
            if xmlkind == 'attribute':
                element = self.create_element(attributes_parent, "attribute", name=field_name)
            else:
                element = self.create_element(parent, "element", name=field_name)
            if field_doc:
                annotation = self.create_element(element, "annotation")
                documentation = self.create_element(annotation, "documentation")
                documentation.text = field_doc
            self.set_field_type(schema_root, record_name, element, field_type)
        
        return element
    
    def create_record(self, schema_root: Element, record: dict) -> str:
        """Convert an Avro record to an XML complex type."""
        name = record['name']
        doc = record.get('doc', '')
        if name in self.known_types:
            return name
        complex_type = self.create_complex_type(schema_root, name=name)
        if doc:
            annotation = self.create_element(complex_type, "annotation")
            documentation = self.create_element(annotation, "documentation")
            documentation.text = doc
        sequence = self.create_element(complex_type, "sequence")
        attributes_parent = complex_type  # Attributes should be direct children of the complexType, not inside the sequence
        for field in record['fields']:
            self.create_field(schema_root, name, sequence, field, attributes_parent)
        self.known_types.append(name)
        return name

    def xsd_namespace_from_avro_namespace(self, namespace: str):
        """Convert an Avro namespace to an XML schema namespace."""
        if not self.target_namespace:
            return "urn:"+namespace.replace('.', ':')
        else:
            return self.target_namespace

    def avro_schema_to_xsd(self, avro_schema: dict) -> Element:
        """Convert the top-level Avro schema to an XML schema."""
        ET.register_namespace('xs', self.xmlns['xs'])
        schema = Element(f"{{{self.xmlns['xs']}}}schema")
        if isinstance(avro_schema, list):
            for record in avro_schema:
                if record['type'] == 'record':
                    if 'namespace' in record:
                        self.update_common_namespace(record['namespace'])
                    self.create_record(schema, record)
                elif record['type'] == 'enum':
                    if 'namespace' in record:
                        self.update_common_namespace(record['namespace'])
                    self.create_enum(schema, record)
        else:
            if avro_schema['type'] == 'record':
                if 'namespace' in avro_schema:
                    self.update_common_namespace(avro_schema['namespace'])
                self.create_record(schema, avro_schema)
            elif avro_schema['type'] == 'enum':
                if 'namespace' in avro_schema:
                    self.update_common_namespace(avro_schema['namespace'])
                self.create_enum(schema, avro_schema)
            elif avro_schema['type'] == 'fixed':
                if 'namespace' in avro_schema:
                    self.update_common_namespace(avro_schema['namespace'])
                self.create_fixed(schema, avro_schema)
        schema.set('targetNamespace', self.xsd_namespace_from_avro_namespace(self.common_namespace))
        schema.set('xmlns', self.xsd_namespace_from_avro_namespace(self.common_namespace))
        ET.register_namespace('', self.xsd_namespace_from_avro_namespace(self.common_namespace))
        return schema

    def save_xsd_to_file(self, schema: Element, xml_path: str) -> None:
        """Save the XML schema to a file."""
        tree_str = tostring(schema, 'utf-8')
        pretty_tree = minidom.parseString(tree_str).toprettyxml(indent="  ")
        with open(xml_path, 'w', encoding='utf-8') as xml_file:
            xml_file.write(pretty_tree)

    def convert_avro_to_xsd(self, avro_schema_path: str, xml_file_path: str) -> None:
        """Convert Avro schema file to XML schema file."""
        with open(avro_schema_path, 'r', encoding='utf-8') as avro_file:
            avro_schema = json.load(avro_file)
            
        xml_schema = self.avro_schema_to_xsd(avro_schema)
        self.save_xsd_to_file(xml_schema, xml_file_path)

def convert_avro_to_xsd(avro_schema_path: str, xml_file_path: str, target_namespace: str = '') -> None:
    avrotoxml = AvroToXSD(target_namespace)
    avrotoxml.convert_avro_to_xsd(avro_schema_path, xml_file_path)



"""
JSON Structure to CDDL (Concise Data Definition Language) converter.

CDDL is defined in RFC 8610 and is a schema language primarily used for
expressing CBOR and JSON data structures.

RFC 8610 Compliance Summary:
============================

Generated CDDL Features:
- Primitive types: uint, int, float16/32/64, tstr, bstr, bool, nil, any
- Maps/objects with member keys and values
- Arrays with homogeneous and heterogeneous (tuple) items
- Choice/union types (type1 / type2)
- Occurrence indicators: ? (optional), * (zero-or-more), + (one-or-more)
- Literal values (strings, integers, floats, booleans)
- Range constraints (min..max)
- Type references
- Groups and group composition
- Comments from descriptions
- Unwrap operator (~) for $extends references

Control Operators Generated:
- .size: From minLength/maxLength for strings, minItems/maxItems for arrays
- .regexp: From pattern validation
- .default: From default value
- .lt, .le, .gt, .ge: From minimum/maximum/exclusiveMinimum/exclusiveMaximum

Notes on Type Mapping:
- JSON Structure binary maps to bstr
- JSON Structure string maps to tstr
- Integer types map to int or uint based on constraints
- Float types map to float16/32/64 based on precision
"""

import json
import logging
import os
from typing import Any, Dict, List, Optional, Set, Tuple, Union

from avrotize.common import avro_name


# Configure module logger
logger = logging.getLogger(__name__)

# Default indentation for CDDL output
INDENT = '  '

# Maximum recursion depth for type conversion (prevents stack overflow)
MAX_CONVERSION_DEPTH = 100


class StructureToCddlError(Exception):
    """
    Exception raised when JSON Structure to CDDL conversion fails.

    Attributes:
        message: Human-readable error description
        context: Optional context about where the error occurred
        cause: Optional underlying exception that caused this error
    """

    def __init__(self, message: str, context: Optional[str] = None,
                 cause: Optional[Exception] = None) -> None:
        self.message = message
        self.context = context
        self.cause = cause
        full_message = message
        if context:
            full_message = f"{message} (context: {context})"
        super().__init__(full_message)


class StructureToCddlCycleError(StructureToCddlError):
    """
    Exception raised when a circular type reference is detected.

    Attributes:
        cycle_path: List of type names forming the cycle
    """

    def __init__(self, cycle_path: List[str]) -> None:
        self.cycle_path = cycle_path
        cycle_str = ' -> '.join(cycle_path)
        super().__init__(f"Circular type reference detected: {cycle_str}")


# JSON Structure primitive types to CDDL type mapping
STRUCTURE_TO_CDDL_TYPES: Dict[str, str] = {
    # Integer types
    'int8': 'int',
    'uint8': 'uint',
    'int16': 'int',
    'uint16': 'uint',
    'int32': 'int',
    'uint32': 'uint',
    'int64': 'int',
    'uint64': 'uint',
    'int128': 'int',
    'uint128': 'uint',
    'integer': 'int',

    # Floating-point types
    'float': 'float32',
    'float8': 'float16',
    'float16': 'float16',
    'float32': 'float32',
    'float64': 'float64',
    'double': 'float64',
    'binary32': 'float32',
    'binary64': 'float64',
    'decimal': 'float64',  # No direct CDDL equivalent
    'number': 'float64',

    # String types
    'string': 'tstr',
    'binary': 'bstr',
    'bytes': 'bstr',

    # Boolean and null
    'boolean': 'bool',
    'null': 'nil',

    # Any type
    'any': 'any',

    # Date/time types (no direct CDDL equivalents, use tstr)
    'date': 'tstr',
    'time': 'tstr',
    'datetime': 'tstr',
    'timestamp': 'tstr',
    'duration': 'tstr',

    # Other types
    'uuid': 'tstr',
    'uri': 'tstr',
    'jsonpointer': 'tstr',
}


class StructureToCddlConverter:
    """
    Converts JSON Structure schema documents to CDDL format.
    """

    def __init__(self) -> None:
        self.definitions: Dict[str, Dict[str, Any]] = {}
        self.schema_registry: Dict[str, Dict] = {}
        self.converted_types: Set[str] = set()
        self.output_lines: List[str] = []
        self._conversion_stack: List[str] = []
        self._conversion_depth: int = 0

    def _to_cddl_name(self, name: str) -> str:
        """
        Convert a name to a valid CDDL identifier.
        
        CDDL identifiers can contain hyphens, so we prefer hyphenated names.
        If the name has altnames.cddl, use that instead.
        """
        # Replace underscores with hyphens for CDDL style
        return name.replace('_', '-')

    def _get_original_name(self, schema: Dict[str, Any], default_name: str) -> str:
        """Get the original CDDL name from altnames if available."""
        altnames = schema.get('altnames', {})
        if isinstance(altnames, dict) and 'cddl' in altnames:
            return altnames['cddl']
        return self._to_cddl_name(default_name)

    def _format_comment(self, description: Optional[str], indent_level: int = 0) -> str:
        """Format a description as a CDDL comment."""
        if not description:
            return ''
        
        indent = INDENT * indent_level
        lines = description.split('\n')
        return '\n'.join(f"{indent}; {line.strip()}" for line in lines if line.strip())

    def convert_structure_to_cddl(self, structure_content: str) -> str:
        """
        Convert JSON Structure content to CDDL format.
        
        Args:
            structure_content: The JSON Structure schema as a string
            
        Returns:
            CDDL schema as a string
        """
        schema = json.loads(structure_content)
        return self.convert_structure_schema_to_cddl(schema)

    def convert_structure_schema_to_cddl(self, schema: Dict[str, Any]) -> str:
        """
        Convert a JSON Structure schema dict to CDDL format.
        
        Args:
            schema: The JSON Structure schema as a dict
            
        Returns:
            CDDL schema as a string
        """
        # Clear state
        self.definitions = schema.get('definitions', {})
        self.schema_registry.clear()
        self.converted_types.clear()
        self.output_lines.clear()
        self._conversion_stack.clear()
        self._conversion_depth = 0

        # Add header comment
        schema_id = schema.get('$id', '')
        if schema_id:
            self.output_lines.append(f"; Generated from {schema_id}")
            self.output_lines.append("")

        # Process root type if present
        root_name = schema.get('name')
        root_type = schema.get('type')
        
        if root_name and root_type:
            # Root schema has inline type definition
            self._convert_definition(root_name, schema)

        # Process all definitions
        for def_name, def_schema in self.definitions.items():
            if def_name not in self.converted_types:
                self._convert_definition(def_name, def_schema)

        return '\n'.join(self.output_lines)

    def _convert_definition(self, name: str, schema: Dict[str, Any]) -> None:
        """
        Convert a single type definition to CDDL.
        
        Args:
            name: The name of the type
            schema: The JSON Structure schema for the type
        """
        if name in self.converted_types:
            return
        
        self.converted_types.add(name)
        
        # Get original CDDL name
        cddl_name = self._get_original_name(schema, name)
        
        # Add description as comment
        description = schema.get('description')
        if description:
            comment = self._format_comment(description)
            if comment:
                self.output_lines.append(comment)
        
        # Convert the type
        type_def = self._convert_type(schema, cddl_name)
        
        # Write the rule
        self.output_lines.append(f"{cddl_name} = {type_def}")
        self.output_lines.append("")

    def _convert_type(self, schema: Dict[str, Any], context_name: str = '') -> str:
        """
        Convert a JSON Structure type to CDDL type expression.
        
        Args:
            schema: The JSON Structure type schema
            context_name: Name context for nested types
            
        Returns:
            CDDL type expression as a string
        """
        if not isinstance(schema, dict):
            if isinstance(schema, str):
                # Simple type reference
                return self._map_primitive_type(schema)
            return 'any'

        # Check recursion depth
        self._conversion_depth += 1
        if self._conversion_depth > MAX_CONVERSION_DEPTH:
            self._conversion_depth -= 1
            logger.warning("Maximum conversion depth exceeded for context: %s", context_name)
            raise StructureToCddlError(
                f"Maximum conversion depth ({MAX_CONVERSION_DEPTH}) exceeded",
                context=context_name
            )

        try:
            # Handle $ref
            if '$ref' in schema:
                return self._convert_ref(schema['$ref'])

            # Check for enum first (can have both type and enum)
            if 'enum' in schema:
                return self._convert_enum(schema['enum'])

            type_value = schema.get('type')
            
            # Handle union types (type is a list)
            if isinstance(type_value, list):
                return self._convert_union(type_value, schema, context_name)

            # Handle specific types
            if type_value == 'object':
                return self._convert_object(schema, context_name)
            elif type_value == 'array':
                return self._convert_array(schema, context_name)
            elif type_value == 'tuple':
                return self._convert_tuple(schema, context_name)
            elif type_value == 'map':
                return self._convert_map(schema, context_name)
            elif type_value is not None:
                return self._convert_primitive(type_value, schema)
            
            # Check for const
            if 'const' in schema:
                return self._convert_const(schema['const'])

            return 'any'
        finally:
            self._conversion_depth -= 1

    def _map_primitive_type(self, type_name: str) -> str:
        """Map a JSON Structure primitive type to CDDL type."""
        return STRUCTURE_TO_CDDL_TYPES.get(type_name, type_name)

    def _convert_ref(self, ref: str) -> str:
        """Convert a $ref to a CDDL type reference."""
        if ref.startswith('#/definitions/'):
            type_name = ref[len('#/definitions/'):]
            # Get original name if available
            if type_name in self.definitions:
                return self._get_original_name(self.definitions[type_name], type_name)
            return self._to_cddl_name(type_name)
        elif ref.startswith('#/'):
            # Local reference
            return self._to_cddl_name(ref.split('/')[-1])
        else:
            # External reference - use as-is
            return self._to_cddl_name(ref)

    def _convert_union(self, types: List[Any], schema: Dict[str, Any], context_name: str) -> str:
        """Convert a union type to CDDL choice."""
        cddl_types = []
        for t in types:
            if isinstance(t, str):
                cddl_types.append(self._map_primitive_type(t))
            elif isinstance(t, dict):
                cddl_types.append(self._convert_type(t, context_name))
            else:
                cddl_types.append(str(t))
        
        return ' / '.join(cddl_types)

    def _convert_object(self, schema: Dict[str, Any], context_name: str) -> str:
        """Convert an object type to CDDL map."""
        properties = schema.get('properties', {})
        required = set(schema.get('required', []))
        extends = schema.get('$extends')
        
        if not properties and not extends:
            # Empty object
            return '{ }'

        lines = ['{']
        
        # Handle $extends (unwrap operator in CDDL)
        if extends:
            extends_refs = extends if isinstance(extends, list) else [extends]
            for ext_ref in extends_refs:
                if isinstance(ext_ref, str):
                    ref_name = self._convert_ref(ext_ref)
                elif isinstance(ext_ref, dict) and '$ref' in ext_ref:
                    ref_name = self._convert_ref(ext_ref['$ref'])
                else:
                    continue
                lines.append(f"{INDENT}~{ref_name},")

        # Convert properties
        prop_lines = []
        for prop_name, prop_schema in properties.items():
            original_name = self._get_original_name(prop_schema, prop_name) if isinstance(prop_schema, dict) else self._to_cddl_name(prop_name)
            prop_type = self._convert_type(prop_schema, f"{context_name}.{prop_name}")
            
            # Add optional marker if not required
            optional_marker = '' if prop_name in required else '? '
            
            prop_lines.append(f"{INDENT}{optional_marker}{original_name}: {prop_type}")

        lines.extend([f"{line}," for line in prop_lines[:-1]] if len(prop_lines) > 1 else [])
        if prop_lines:
            lines.append(f"{INDENT}{prop_lines[-1].strip()}" if len(prop_lines) == 1 else prop_lines[-1])
        
        lines.append('}')
        
        return '\n'.join(lines)

    def _convert_array(self, schema: Dict[str, Any], context_name: str) -> str:
        """Convert an array type to CDDL array."""
        items = schema.get('items', {'type': 'any'})
        items_type = self._convert_type(items, f"{context_name}[]")
        
        # Get cardinality constraints
        min_items = schema.get('minItems')
        max_items = schema.get('maxItems')
        
        if min_items is not None and max_items is not None:
            if min_items == max_items:
                return f"[{min_items}*{min_items} {items_type}]"
            else:
                return f"[{min_items}*{max_items} {items_type}]"
        elif min_items is not None and min_items > 0:
            if min_items == 1:
                return f"[+ {items_type}]"
            return f"[{min_items}* {items_type}]"
        elif max_items is not None:
            return f"[*{max_items} {items_type}]"
        else:
            return f"[* {items_type}]"

    def _convert_tuple(self, schema: Dict[str, Any], context_name: str) -> str:
        """Convert a tuple type to CDDL array with positional elements."""
        tuple_order = schema.get('tuple', [])
        properties = schema.get('properties', {})
        
        if not tuple_order:
            return '[ ]'

        elements = []
        for idx, prop_name in enumerate(tuple_order):
            prop_schema = properties.get(prop_name, {'type': 'any'})
            elem_type = self._convert_type(prop_schema, f"{context_name}[{idx}]")
            elements.append(elem_type)
        
        return f"[{', '.join(elements)}]"

    def _convert_map(self, schema: Dict[str, Any], context_name: str) -> str:
        """Convert a map type to CDDL map with computed key."""
        keys = schema.get('keys', {'type': 'string'})
        values = schema.get('values', {'type': 'any'})
        
        key_type = self._convert_type(keys, f"{context_name}.keys")
        value_type = self._convert_type(values, f"{context_name}.values")
        
        return f"{{ * {key_type} => {value_type} }}"

    def _convert_primitive(self, type_name: str, schema: Dict[str, Any]) -> str:
        """Convert a primitive type with optional constraints."""
        cddl_type = self._map_primitive_type(type_name)
        
        # Check for numeric range - use CDDL range syntax instead of type + control operators
        minimum = schema.get('minimum')
        maximum = schema.get('maximum')
        exclusive_min = schema.get('exclusiveMinimum')
        exclusive_max = schema.get('exclusiveMaximum')
        
        # If we have both bounds, use range syntax
        if minimum is not None and maximum is not None:
            return f"{minimum}..{maximum}"
        elif exclusive_min is not None and exclusive_max is not None:
            return f"{exclusive_min + 1}..{exclusive_max - 1}"
        elif minimum is not None and exclusive_max is not None:
            return f"{minimum}..{exclusive_max - 1}"
        elif exclusive_min is not None and maximum is not None:
            return f"{exclusive_min + 1}..{maximum}"
        
        # For single-bound constraints or other constraints, use control operators
        constraints = self._get_constraints(schema)
        if constraints:
            return f"{cddl_type} {constraints}"
        return cddl_type

    def _get_constraints(self, schema: Dict[str, Any]) -> str:
        """
        Get CDDL control operators for constraints.
        
        Note: CDDL only allows ONE control operator per type. If multiple are needed,
        we prioritize: pattern > size > numeric > default > const
        """
        # String constraints
        if 'minLength' in schema or 'maxLength' in schema:
            min_len = schema.get('minLength', 0)
            max_len = schema.get('maxLength')
            if min_len == max_len and max_len is not None:
                return f".size {max_len}"
            elif max_len is not None:
                return f".size ({min_len}..{max_len})"
            elif min_len > 0:
                return f".size ({min_len}..)"
        
        # Pattern constraint (takes priority for strings)
        if 'pattern' in schema:
            pattern = schema['pattern'].replace('"', '\\"')
            return f'.regexp "{pattern}"'
        
        # Numeric constraints (only if not already handled by range syntax)
        minimum = schema.get('minimum')
        maximum = schema.get('maximum')
        exclusive_min = schema.get('exclusiveMinimum')
        exclusive_max = schema.get('exclusiveMaximum')
        
        # Single-bound numeric constraints
        if minimum is not None and maximum is None and exclusive_max is None:
            return f".ge {minimum}"
        if maximum is not None and minimum is None and exclusive_min is None:
            return f".le {maximum}"
        if exclusive_min is not None and exclusive_max is None and maximum is None:
            return f".gt {exclusive_min}"
        if exclusive_max is not None and exclusive_min is None and minimum is None:
            return f".lt {exclusive_max}"
        
        # Default value
        if 'default' in schema:
            default = schema['default']
            if isinstance(default, str):
                return f'.default "{default}"'
            elif isinstance(default, bool):
                return f".default {'true' if default else 'false'}"
            elif default is None:
                return ".default nil"
            else:
                return f".default {default}"
        
        # Const value (use .eq)
        if 'const' in schema:
            const = schema['const']
            if isinstance(const, str):
                return f'.eq "{const}"'
            elif isinstance(const, bool):
                return f".eq {'true' if const else 'false'}"
            elif const is None:
                return ".eq nil"
            else:
                return f".eq {const}"

        return ''

    def _convert_const(self, value: Any) -> str:
        """Convert a const value to CDDL literal."""
        if isinstance(value, str):
            return f'"{value}"'
        elif isinstance(value, bool):
            return 'true' if value else 'false'
        elif value is None:
            return 'nil'
        elif isinstance(value, (int, float)):
            return str(value)
        else:
            return f'"{value}"'

    def _convert_enum(self, values: List[Any]) -> str:
        """Convert an enum to CDDL choice of literals."""
        cddl_values = [self._convert_const(v) for v in values]
        return ' / '.join(cddl_values)


def convert_structure_to_cddl(structure_content: str) -> str:
    """
    Convert JSON Structure content to CDDL.
    
    Args:
        structure_content: JSON Structure schema as a string
        
    Returns:
        CDDL schema as a string
    """
    converter = StructureToCddlConverter()
    return converter.convert_structure_to_cddl(structure_content)


def convert_structure_to_cddl_files(
    structure_schema_path: str,
    cddl_file_path: Optional[str] = None
) -> str:
    """
    Convert a JSON Structure schema file to CDDL.
    
    Args:
        structure_schema_path: Path to the JSON Structure schema file
        cddl_file_path: Optional path for the output CDDL file
        
    Returns:
        CDDL schema as a string
    """
    with open(structure_schema_path, 'r', encoding='utf-8') as f:
        structure_content = f.read()
    
    cddl_content = convert_structure_to_cddl(structure_content)
    
    if cddl_file_path:
        os.makedirs(os.path.dirname(cddl_file_path) or '.', exist_ok=True)
        with open(cddl_file_path, 'w', encoding='utf-8') as f:
            f.write(cddl_content)
    
    return cddl_content



import copy
import json
import argparse
import os
from typing import Literal, NamedTuple, Dict, Any, List

indent = '  '

Comment = NamedTuple('Comment', [('content', str), ('tags', Dict[str, Any])])
Oneof = NamedTuple('Oneof', [('comment', 'Comment'), ('name', str), ('fields', List['Field'])])
Field = NamedTuple('Field', [('comment', 'Comment'), ('label', str), ('type', str), ('key_type', str), ('val_type', str), ('name', str), ('number', int), ('dependencies', List[str])])
Enum = NamedTuple('Enum', [('comment', 'Comment'), ('name', str), ('fields', Dict[str, 'Field'])])
Message = NamedTuple('Message', [('comment', 'Comment'), ('name', str), ('fields', List['Field']), ('oneofs', List['Oneof']),
                                        ('messages', Dict[str, 'Message']), ('enums', Dict[str, 'Enum']), ('dependencies', List[str])])
Service = NamedTuple('Service', [('name', str), ('functions', Dict[str, 'RpcFunc'])])
RpcFunc = NamedTuple('RpcFunc', [('name', str), ('in_type', str), ('out_type', str), ('uri', str)])
ProtoFile = NamedTuple('ProtoFile',
                            [('messages', Dict[str, 'Message']), ('enums', Dict[str, 'Enum']),
                            ('services', Dict[str, 'Service']), ('imports', List[str]),
                            ('options', Dict[str, str]), ('package', str)])
ProtoFiles = NamedTuple('ProtoFiles', [('files', List['ProtoFile'])])

class AvroToProto:
    
    def __init__(self) -> None:
        self.naming_mode: Literal['snake', 'pascal', 'camel'] = 'pascal'
        self.allow_optional: bool = False
        self.default_namespace: str = ''

    def avro_primitive_to_proto_type(self, avro_type: str, dependencies: List[str]) -> str:
        """Map Avro primitive types to Protobuf types."""
        mapping = {
            'null': 'google.protobuf.Empty',  # Special handling may be required
            'boolean': 'bool',
            'int': 'int32',
            'long': 'int64',
            'float': 'float',
            'double': 'double',
            'bytes': 'bytes',
            'string': 'string',
        }
        # logical types require special handling
        if isinstance(avro_type, dict) and 'logicalType' in avro_type:
            logical_type = avro_type['logicalType']
            if logical_type == 'date':
                return 'string'
            elif logical_type == 'time-millis':
                return 'string'
            elif logical_type == 'timestamp-millis':
                return 'string'
            elif logical_type == 'decimal':
                precision = avro_type['precision']
                scale = avro_type['scale']
                return 'string'
            elif logical_type == 'duration':
                return 'string'
            elif logical_type == 'uuid':
                return 'string'
            
        type = mapping.get(avro_type, '')
        if not type:
            dependencies.append(avro_type)
            type = avro_type
        return type

    def compose_name(self, prefix: str, name: str, naming_mode: Literal['pascal', 'camel', 'snake', 'default', 'field'] = 'default') -> str:
        if naming_mode == 'default':
            naming_mode = self.naming_mode
        if naming_mode == 'field':
            if self.naming_mode == 'pascal':
                naming_mode = 'camel'
            else:
                naming_mode = self.naming_mode
        if naming_mode == 'snake':
            return f"{prefix}_{name}"
        if naming_mode == 'pascal':
            return f"{prefix[0].upper()+prefix[1:] if prefix else ''}{name[0].upper()+name[1:] if name else ''}"
        if naming_mode == 'camel':
            return f"{prefix[0].lower()+prefix[1:] if prefix else ''}{name[0].upper()+name[1:] if name else ''}"
        return prefix+name    

    def convert_field(self, message: Message, avro_field: dict, index: int, proto_files: ProtoFiles) -> Field | Oneof | Enum | Message:
        """Convert an Avro field to a Protobuf field."""
        field_type = avro_field['type']
        field_name = avro_field['name'] if 'name' in avro_field else self.compose_name(field_type.split('.')[-1],'value', 'field') if isinstance(field_type, str) else self.compose_name(f"_{index}", 'value', 'field')
        if 'doc' in avro_field:
            comment = Comment(avro_field["doc"], {})
        else:
            comment = Comment('',{})
        
        return self.convert_field_type(message, field_name, field_type, comment, index, proto_files)
        
    def convert_record_type(self, avro_record: dict, comment: Comment, proto_files: ProtoFiles) -> Message:
        """Convert an Avro record to a Protobuf message."""
        local_message = Message(comment, avro_record['name'], [], [], {}, {}, [])
        offs = 1
        for i, f in enumerate(avro_record['fields']):
            field = self.convert_field(local_message, f, i+offs, proto_files)
            if isinstance(field, Oneof):
                for f in field.fields:
                    local_message.dependencies.extend(f.dependencies)
                local_message.oneofs.append(field)
                offs += len(field.fields)-1
            elif isinstance(field, Enum):
                enum = Enum(field.comment, self.compose_name(field.name,'enum'), field.fields)
                local_message.enums[enum.name] = enum
                local_message.fields.append(Field(field.comment, '', enum.name, '', '', field.name.split('.')[-1], i+offs, []))
            elif isinstance(field, Message):
                inner_message = Message(field.comment, self.compose_name(field.name,'type'), field.fields, field.oneofs, field.messages, field.enums, [])
                local_message.messages[inner_message.name] = inner_message
                local_message.fields.append(Field(field.comment, '', inner_message.name, '', '', field.name.split('.')[-1], i+offs, []))
                local_message.dependencies.extend(field.dependencies)
            else:
                local_message.dependencies.extend(field.dependencies)
                local_message.fields.append(field)
        return local_message

    def convert_field_type(self, message: Message, field_name: str, field_type: str | dict | list, comment: Comment, index: int, proto_files: ProtoFiles) -> Field | Oneof | Enum | Message:
        """Convert an Avro field type to a Protobuf field type."""
        label = ''
        
        if isinstance(field_type, list):
            # Handling union types (including nullable fields)
            non_null_types = [t for t in field_type if t != 'null']
            if len(non_null_types) == 1:
                if self.allow_optional:
                    label = 'optional'
                field_type = non_null_types[0]
            elif len(non_null_types) > 0:
                oneof_fields = []
                for i, t in enumerate(non_null_types):
                    field = self.convert_field_type(message, self.compose_name(field_name,'choice', 'field'), t, comment, i+index, proto_files)
                    if isinstance(field, Field):
                        if field.type == 'map' or field.type == 'array':
                            local_message = Message(comment, self.compose_name(field.name,field.type), [], [], {}, {}, field.dependencies)
                            local_message.fields.append(field)
                            new_field = Field(field.comment, '', local_message.name, '', '', self.compose_name(field.name.split('.')[-1],field.type, 'field'), i+index, field.dependencies)
                            message.messages[local_message.name] = local_message
                            oneof_fields.append(new_field)
                        else:
                            field = Field(field.comment, field.label, field.type, field.key_type, field.val_type, self.compose_name(field_name, (field.type.split('.')[-1]), 'field'), i+index, field.dependencies)
                            oneof_fields.append(field)
                    elif isinstance(field, Oneof):
                        deps: List[str] = []
                        oneof = field
                        for f in oneof.fields:
                            deps.extend(f.dependencies)
                        local_message = Message(comment, self.compose_name(field.name,'choice'), [], [], {}, {}, deps)
                        index += len(field.fields)
                        local_message.oneofs.append(field)
                        new_field = Field(field.comment, '', local_message.name, '', '', field.name.split('.')[-1], i+index, deps)
                        message.messages[local_message.name] = local_message
                        oneof_fields.append(new_field)
                    elif isinstance(field, Enum):
                        enum = Enum(field.comment, self.compose_name(field.name,"options"), field.fields)
                        message.enums[enum.name] = enum
                        field = Field(field.comment, '', enum.name, '', '', field.name.split('.')[-1], i+index, [])
                        oneof_fields.append(field)
                    elif isinstance(field, Message):
                        local_message = Message(field.comment, self.compose_name(field.name,'type'), field.fields, field.oneofs, field.messages, field.enums, field.dependencies)
                        message.messages[local_message.name] = local_message
                        field = Field(field.comment, '', local_message.name, '', '', field.name.split('.')[-1], i+index, field.dependencies)
                        oneof_fields.append(field)
                oneof = Oneof(comment, field_name, copy.deepcopy(oneof_fields))
                return oneof
            else:
                raise ValueError(f"Field {field_name} is a union type without any non-null types")

        if isinstance(field_type, dict):
            # Nested types (e.g., records, enums) require special handling
            if field_type['type'] == 'record':
                return self.convert_record_type(field_type, comment, proto_files)
            elif field_type['type'] == 'enum':
                enum_symbols = {symbol: Field(comment, '', symbol, '', '', symbol, s, []) for s, symbol in enumerate(field_type['symbols'])}
                return Enum(comment, field_type['name'], enum_symbols)
            elif field_type['type'] == 'array':
                converted_field_type = self.convert_field_type(message, self.compose_name(field_name, "item"), field_type['items'], comment, index, proto_files)
                if isinstance(converted_field_type, Field):
                    return Field(comment, 'repeated', 'array', '', converted_field_type.type, field_name, index, converted_field_type.dependencies)
                elif isinstance(converted_field_type, Enum):
                    enum = Enum(converted_field_type.comment, self.compose_name(converted_field_type.name,'enum'), converted_field_type.fields)
                    message.enums[enum.name] = enum
                    return Field(comment, 'repeated', 'array', '', enum.name, field_name, index, [])
                elif isinstance(converted_field_type, Message):
                    local_message = Message(converted_field_type.comment, self.compose_name(converted_field_type.name,'type'), converted_field_type.fields, converted_field_type.oneofs, converted_field_type.messages, converted_field_type.enums, converted_field_type.dependencies)
                    message.messages[local_message.name] = local_message
                    return Field(comment, 'repeated', 'array', '', local_message.name, field_name, index, [])
                elif isinstance(converted_field_type, Oneof):
                    deps3: List[str] = []
                    fl = []
                    for i, f in enumerate(converted_field_type.fields):
                        fl.append(Field(Comment('',{}), '', f.type, '', '', f.name, i+1, []))
                        deps3.extend(f.dependencies)
                    oneof = Oneof(converted_field_type.comment, 'item', fl)
                    local_message = Message(comment, self.compose_name(field_name,'type'), [], [], {}, {}, deps3)
                    local_message.oneofs.append(oneof)
                    new_field = Field(Comment('',{}), 'repeated', 'array', '', local_message.name, field_name.split('.')[-1], index, local_message.dependencies)
                    message.messages[local_message.name] = local_message
                    return new_field
            elif field_type['type'] == 'map':
                converted_field_type = self.convert_field_type(message, self.compose_name(field_name,'item', 'field'), field_type['values'], comment, index, proto_files)
                if isinstance(converted_field_type, Field):
                    return Field(comment, label, 'map', 'string', converted_field_type.type, field_name, index, converted_field_type.dependencies)
                elif isinstance(converted_field_type, Enum):
                    enum = Enum(converted_field_type.comment, self.compose_name(converted_field_type.name,'enum'), converted_field_type.fields)
                    message.enums[enum.name] = enum
                    return Field(comment, label, 'map', 'string', enum.name, field_name, index, [])
                elif isinstance(converted_field_type, Message):
                    local_message = Message(converted_field_type.comment, self.compose_name(converted_field_type.name,'type'), converted_field_type.fields, converted_field_type.oneofs, converted_field_type.messages, converted_field_type.enums, [])
                    message.messages[local_message.name] = local_message
                    return Field(comment, label, 'map', 'string', local_message.name, field_name, index, local_message.dependencies)
                elif isinstance(converted_field_type, Oneof):
                    deps4: List[str] = []
                    fl = []
                    for i, f in enumerate(converted_field_type.fields):
                        fl.append(Field(Comment('',{}), '', f.type, '', '', f.name, i+1, []))
                        deps4.extend(f.dependencies)
                    oneof = Oneof(converted_field_type.comment, 'item', fl)
                    local_message = Message(comment, self.compose_name(field_name, 'type'), [], [], {}, {}, deps4)
                    local_message.oneofs.append(oneof)
                    new_field = Field(Comment('',{}), label, 'map', 'string', local_message.name, field_name.split('.')[-1], index, local_message.dependencies)
                    message.messages[local_message.name] = local_message
                    return new_field
            elif field_type['type'] == "fixed":
                return Field(comment, label, 'fixed','string', 'string', field_name, index, [])
            else:
                deps1: List[str] = []
                proto_type = self.avro_primitive_to_proto_type(field_type['type'], deps1)
                return Field(comment, label, proto_type, '', '', field_name, index, deps1)
        elif isinstance(field_type, str):
            deps2: List[str] = []
            proto_type = self.avro_primitive_to_proto_type(field_type, deps2)
            return Field(comment, label, proto_type, '', '', field_name, index, deps2)
        raise ValueError(f"Unknown field type {field_type}")

    def avro_schema_to_proto_message(self, avro_schema: dict, proto_files: ProtoFiles) -> str:
        """Convert an Avro schema to a Protobuf message definition."""
        comment = Comment('',{})
        if 'doc' in avro_schema:
            comment = Comment(avro_schema["doc"], {})
        namespace = avro_schema.get("namespace", '')
        if not namespace:
            namespace = self.default_namespace
        if avro_schema['type'] == 'record':
            message = self.convert_record_type(avro_schema, comment, proto_files)
            file = next((f for f in proto_files.files if f.package == namespace), None)
            if not file:
                file = ProtoFile({}, {}, {}, [], {}, namespace)
                proto_files.files.append(file)
            file.messages[message.name] = message
        elif avro_schema['type'] == 'enum':
            enum_name = avro_schema['name']
            enum_symbols = {symbol: Field(comment, '', symbol, '', '', symbol, s, []) for s, symbol in enumerate(avro_schema['symbols'])}
            enum = Enum(comment, enum_name, enum_symbols)
            file = next((f for f in proto_files.files if f.package == namespace), None)
            if not file:
                file = ProtoFile({}, {}, {}, [], {}, namespace)
                proto_files.files.append(file)
            file.enums[enum_name] = enum
        return avro_schema["name"]
        
    def avro_schema_to_proto_messages(self, avro_schema_input, proto_files: ProtoFiles):
        """Convert an Avro schema to Protobuf message definitions."""
        if not isinstance(avro_schema_input, list):
            avro_schema_list = [avro_schema_input]
        else:
            avro_schema_list = avro_schema_input    
        for avro_schema in avro_schema_list:
            self.avro_schema_to_proto_message(avro_schema, proto_files)

    def save_proto_to_file(self, proto_files: ProtoFiles, proto_path):
        """Save the Protobuf schema to a file."""
        for proto in proto_files.files:
            # gather dependencies that are within the package
            deps: List[str] = []
            for message in proto.messages.values():
                for dep in message.dependencies:
                    if '.' in dep:
                        deps.append(dep.rsplit('.',1)[0])
            deps = list(set(deps))

            #proto.imports.extend([f.package[len(proto.package)+1:] for f in proto_files.files if f.package.startswith(proto.package) and f.package != proto.package])
            proto.imports.extend([d for d in deps if d != proto.package])
            proto_file_path = os.path.join(proto_path, f"{proto.package}.proto")
            # create the directory for the proto file if it doesn't exist
            proto_dir = os.path.dirname(proto_file_path)
            if not os.path.exists(proto_dir):
                os.makedirs(proto_dir, exist_ok=True)
            with open(proto_file_path, 'w') as proto_file:
                # dump the ProtoFile structure in proto syntax
                proto_str = f'syntax = "proto3";\n\n'
                proto_str += f'package {proto.package};\n\n'

                for import_package in proto.imports:
                    proto_str += f"import \"{import_package}.proto\";\n"
                if (len(proto.imports)):
                    proto_str += "\n"
                for enum_name, enum in proto.enums.items():
                    proto_str += f"enum {enum_name} {{\n"
                    for _, field in enum.fields.items():
                        proto_str += f"{indent}{field.name} = {field.number};\n"
                    proto_str += "}\n\n"
                for message in proto.messages.values():
                    proto_str += self.render_message(message)
                for service in proto.services.values():
                    proto_str += f"service {service.name} {{\n"
                    for function_name, func in service.functions.items():
                        proto_str += f"{indent}rpc {func.name} ({func.in_type}) returns ({func.out_type}) {{\n"
                        proto_str += f"{indent}{indent}option (google.api.http) = {{\n"
                        proto_str += f"{indent}{indent}{indent}post: \"{func.uri}\"\n"
                        proto_str += f"{indent}{indent}}};\n"
                        proto_str += f"{indent}}};\n"
                    proto_str += "}\n\n"
                proto_file.write(proto_str)

    def render_message(self, message, level=0) -> str:
        proto_str = f"{indent*level}message {message.name} {{\n"
        fieldsAndOneofs = message.fields+message.oneofs
        fieldsAndOneofs.sort(key=lambda f: f.number if isinstance(f, Field) else f.fields[0].number)
        for fo in fieldsAndOneofs:
            if isinstance(fo, Field):
                field = fo
                if field.type == "map":
                    proto_str += f"{indent*level}{indent}{field.label}{' ' if field.label else ''}map<{field.key_type}, {field.val_type}> {field.name} = {field.number};\n"
                elif field.type == "array":
                    proto_str += f"{indent*level}{indent}{field.label}{' ' if field.label else ''}{field.val_type} {field.name} = {field.number};\n"
                else:
                    proto_str += f"{indent*level}{indent}{field.label}{' ' if field.label else ''}{field.type} {field.name} = {field.number};\n"
            else:
                oneof = fo
                proto_str += f"{indent*level}{indent}oneof {oneof.name} {{\n"
                for field in oneof.fields:
                    proto_str += f"{indent*level}{indent}{indent}{field.label}{' ' if field.label else ''}{field.type} {field.name} = {field.number};\n"
                proto_str += f"{indent*level}{indent}}}\n"
        for enum in message.enums.values():
            proto_str += f"{indent*level}{indent}enum {enum.name} {{\n"
            for _, field in enum.fields.items():
                proto_str += f"{indent*level}{indent}{indent}{field.label}{' ' if field.label else ''}{field.name} = {field.number};\n"
            proto_str += f"{indent*level}{indent}}}\n"
        for local_message in message.messages.values():
            proto_str += self.render_message(local_message, level+1)
        proto_str += f"{indent*level}}}\n"
        return proto_str
            

    def convert_avro_to_proto(self, avro_schema_path, proto_file_path):
        """Convert Avro schema file to Protobuf .proto file."""
        with open(avro_schema_path, 'r') as avro_file:
            avro_schema = json.load(avro_file)
        proto_files = ProtoFiles([])
        self.avro_schema_to_proto_messages(avro_schema, proto_files)
        self.save_proto_to_file(proto_files, proto_file_path)

def convert_avro_to_proto(avro_schema_path, proto_file_path, naming_mode: Literal['snake', 'pascal', 'camel'] = 'pascal', allow_optional: bool = False):
    avrotoproto = AvroToProto()
    avrotoproto.naming_mode = naming_mode
    avrotoproto.allow_optional = allow_optional
    avrotoproto.default_namespace = os.path.splitext(os.path.basename(proto_file_path))[0].replace('-','_')
    avrotoproto.convert_avro_to_proto(avro_schema_path, proto_file_path)